# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2023 Ondsel <development@ondsel.com>                     *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import math

import FreeCAD as App
import Part

from PySide import QtCore
from PySide.QtCore import QT_TRANSLATE_NOOP
from collections.abc import Sequence

if App.GuiUp:
    import FreeCADGui as Gui
    from PySide import QtGui, QtWidgets

__title__ = "Assembly Joint object"
__author__ = "Ondsel"
__url__ = "https://www.freecad.org"

from pivy import coin
import UtilsAssembly
import Preferences

from SoSwitchMarker import SoSwitchMarker

translate = App.Qt.translate

TranslatedJointTypes = [
    translate("Assembly", "Fixed"),
    translate("Assembly", "Revolute"),
    translate("Assembly", "Cylindrical"),
    translate("Assembly", "Slider"),
    translate("Assembly", "Ball"),
    translate("Assembly", "Distance"),
    translate("Assembly", "Parallel"),
    translate("Assembly", "Perpendicular"),
    translate("Assembly", "Angle"),
    translate("Assembly", "RackPinion"),
    translate("Assembly", "Screw"),
    translate("Assembly", "Gears"),
    translate("Assembly", "Belt"),
]

JointTypes = [
    "Fixed",
    "Revolute",
    "Cylindrical",
    "Slider",
    "Ball",
    "Distance",
    "Parallel",
    "Perpendicular",
    "Angle",
    "RackPinion",
    "Screw",
    "Gears",
    "Belt",
]

JointUsingAngle = [
    "Angle",
]

JointUsingDistance = [
    "Distance",
    "RackPinion",
    "Screw",
    "Gears",
    "Belt",
]

JointUsingDistance2 = [
    "Gears",
    "Belt",
]

JointNoNegativeDistance = [
    "Gears",
    "Belt",
]

JointUsingOffset = [
    "Fixed",
    "Revolute",
]

JointUsingRotation = [
    "Fixed",
    "Slider",
]

JointUsingReverse = [
    "Fixed",
    "Revolute",
    "Cylindrical",
    "Slider",
    "Distance",
    "Parallel",
]

JointUsingLimitLength = [
    "Cylindrical",
    "Slider",
]

JointUsingLimitAngle = [
    "Revolute",
    "Cylindrical",
]

JointUsingPreSolve = [
    "Fixed",
    "Revolute",
    "Cylindrical",
    "Slider",
    "Ball",
]

JointParallelForbidden = [
    "Angle",
    "Perpendicular",
]


def solveIfAllowed(assembly, storePrev=False):
    if assembly.Type == "Assembly" and Preferences.preferences().GetBool(
        "SolveInJointCreation", True
    ):
        assembly.solve(storePrev)


def getContext(obj):
    """Fetch the context of an object."""
    context = []
    current = obj

    while current:
        # Add the object's Label at the beginning or the Name if label is empty
        context.insert(0, current.Label if current.Label else current.Name)
        # Get the immediate parent object
        parents = getattr(current, "InList", [])
        current = parents[0] if parents else None
    return ".".join(context)


# The joint object consists of 2 JCS (joint coordinate systems) and a Joint Type.
# A JCS is a placement that is computed (unless it is detached) from references (PropertyXLinkSubHidden) that links to :
# - An object: this can be any Part::Feature solid. Or a PartDesign Body. Or a App::Link to those.
# - An element name: This can be either a face, an edge, a vertex or empty. Empty means that the Object placement will be used
# - A vertex name: For faces and edges, we need to specify which vertex of said face/edge to use
# Both element names hold the full path to the object.
# From these a placement is computed. It is relative to the Object.
class Joint:
    def __init__(self, joint, type_index):
        joint.Proxy = self

        joint.addExtension("App::SuppressibleExtensionPython")

        joint.addProperty(
            "App::PropertyEnumeration",
            "JointType",
            "Joint",
            QT_TRANSLATE_NOOP("App::Property", "The type of the joint"),
            locked=True,
        )
        joint.JointType = JointTypes  # sets the list
        joint.JointType = JointTypes[type_index]  # set the initial value

        self.createProperties(joint)

        self.setJointConnectors(joint, [])

    def onDocumentRestored(self, joint):
        self.createProperties(joint)

    def createProperties(self, joint):
        self.migrationScript(joint)
        self.migrationScript2(joint)
        self.migrationScript3(joint)
        self.migrationScript4(joint)
        self.migrationScript5(joint)
        self.migrationScript6(joint)
        self.migrationScript7(joint)

        # First Joint Connector
        if not hasattr(joint, "Reference1"):
            self.addReference1Property(joint)

        if not hasattr(joint, "Placement1"):
            joint.addProperty(
                "App::PropertyPlacement",
                "Placement1",
                "Joint Connector 1",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "This is the local coordinate system within Reference1's object that will be used for the joint",
                ),
                locked=True,
            )

        if not hasattr(joint, "Detach1"):
            joint.addProperty(
                "App::PropertyBool",
                "Detach1",
                "Joint Connector 1",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "This prevents Placement1 from recomputing, enabling custom positioning of the placement",
                ),
                locked=True,
            )

        if not hasattr(joint, "Offset1"):
            joint.addProperty(
                "App::PropertyPlacement",
                "Offset1",
                "Joint Connector 1",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "This is the attachment offset of the first connector of the joint",
                ),
                locked=True,
            )

        # Second Joint Connector
        if not hasattr(joint, "Reference2"):
            self.addReference2Property(joint)

        if not hasattr(joint, "Placement2"):
            joint.addProperty(
                "App::PropertyPlacement",
                "Placement2",
                "Joint Connector 2",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "This is the local coordinate system within Reference2's object that will be used for the joint",
                ),
                locked=True,
            )

        if not hasattr(joint, "Detach2"):
            joint.addProperty(
                "App::PropertyBool",
                "Detach2",
                "Joint Connector 2",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "This prevents Placement2 from recomputing, enabling custom positioning of the placement",
                ),
                locked=True,
            )

        if not hasattr(joint, "Offset2"):
            joint.addProperty(
                "App::PropertyPlacement",
                "Offset2",
                "Joint Connector 2",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "This is the attachment offset of the second connector of the joint",
                ),
                locked=True,
            )

        # Other properties
        if not hasattr(joint, "Angle"):
            self.addAngleProperty(joint)

        if not hasattr(joint, "Distance"):
            self.addDistanceProperty(joint)

        if not hasattr(joint, "Distance2"):
            self.addDistance2Property(joint)

        if not hasattr(joint, "EnableLengthMin"):
            joint.addProperty(
                "App::PropertyBool",
                "EnableLengthMin",
                "Limits",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Enable the minimum length limit of the joint",
                ),
                locked=True,
            )
            joint.EnableLengthMin = False

        if not hasattr(joint, "EnableLengthMax"):
            joint.addProperty(
                "App::PropertyBool",
                "EnableLengthMax",
                "Limits",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Enable the maximum length limit of the joint",
                ),
                locked=True,
            )
            joint.EnableLengthMax = False

        if not hasattr(joint, "EnableAngleMin"):
            joint.addProperty(
                "App::PropertyBool",
                "EnableAngleMin",
                "Limits",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Enable the minimum angle limit of the joint",
                ),
                locked=True,
            )
            joint.EnableAngleMin = False

        if not hasattr(joint, "EnableAngleMax"):
            joint.addProperty(
                "App::PropertyBool",
                "EnableAngleMax",
                "Limits",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Enable the maximum angle limit of the joint",
                ),
                locked=True,
            )
            joint.EnableAngleMax = False

        if not hasattr(joint, "LengthMin"):
            self.addLengthMinProperty(joint)

        if not hasattr(joint, "LengthMax"):
            self.addLengthMaxProperty(joint)

        if not hasattr(joint, "AngleMin"):
            self.addAngleMinProperty(joint)

        if not hasattr(joint, "AngleMax"):
            self.addAngleMaxProperty(joint)

        joint.setPropertyStatus("Distance", "AllowNegativeValues")
        joint.setPropertyStatus("Distance2", "AllowNegativeValues")
        joint.setPropertyStatus("LengthMin", "AllowNegativeValues")
        joint.setPropertyStatus("LengthMax", "AllowNegativeValues")

    def addReference1Property(self, joint):
        joint.addProperty(
            "App::PropertyXLinkSub",
            "Reference1",
            "Joint Connector 1",
            QT_TRANSLATE_NOOP("App::Property", "The first reference of the joint"),
            locked=True,
        )

    def addReference2Property(self, joint):
        joint.addProperty(
            "App::PropertyXLinkSub",
            "Reference2",
            "Joint Connector 2",
            QT_TRANSLATE_NOOP("App::Property", "The second reference of the joint"),
            locked=True,
        )

    def addAngleProperty(self, joint):
        joint.addProperty(
            "App::PropertyAngle",
            "Angle",
            "Joint",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "This is the angle of the joint. It is used only by the Angle joint.",
            ),
            locked=True,
        )

    def addDistanceProperty(self, joint):
        joint.addProperty(
            "App::PropertyLength",
            "Distance",
            "Joint",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "This is the distance of the joint. It is used only by the Distance joint and Rack and Pinion (pitch radius), Screw and Gears and Belt (radius1)",
            ),
            locked=True,
        )

    def addDistance2Property(self, joint):
        joint.addProperty(
            "App::PropertyLength",
            "Distance2",
            "Joint",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "This is the second distance of the joint. It is used only by the gear joint to store the second radius.",
            ),
            locked=True,
        )

    def addLengthMinProperty(self, joint):
        joint.addProperty(
            "App::PropertyLength",
            "LengthMin",
            "Limits",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "This is the minimum limit for the length between both coordinate systems (along their z-axis)",
            ),
            locked=True,
        )

    def addLengthMaxProperty(self, joint):
        joint.addProperty(
            "App::PropertyLength",
            "LengthMax",
            "Limits",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "This is the maximum limit for the length between both coordinate systems (along their z-axis)",
            ),
            locked=True,
        )

    def addAngleMinProperty(self, joint):
        joint.addProperty(
            "App::PropertyAngle",
            "AngleMin",
            "Limits",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "This is the minimum limit for the angle between both coordinate systems (between their x-axis)",
            ),
            locked=True,
        )

    def addAngleMaxProperty(self, joint):
        joint.addProperty(
            "App::PropertyAngle",
            "AngleMax",
            "Limits",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "This is the maximum limit for the angle between both coordinate systems (between their x-axis)",
            ),
            locked=True,
        )

    def migrationScript(self, joint):
        if hasattr(joint, "Object1") and isinstance(joint.Object1, str):
            objName = joint.Object1
            obj1 = UtilsAssembly.getObjectInPart(objName, joint.Part1)
            el1 = joint.Element1
            vtx1 = joint.Vertex1

            joint.removeProperty("Object1")
            joint.removeProperty("Element1")
            joint.removeProperty("Vertex1")

            joint.addProperty(
                "App::PropertyXLinkSub",
                "Object1",
                "Joint Connector 1",
                QT_TRANSLATE_NOOP("App::Property", "The first object of the joint"),
                locked=True,
            )

            joint.Object1 = [obj1, [el1, vtx1]]

        if hasattr(joint, "Object2") and isinstance(joint.Object2, str):
            objName = joint.Object2
            obj2 = UtilsAssembly.getObjectInPart(objName, joint.Part2)
            el2 = joint.Element2
            vtx2 = joint.Vertex2

            joint.removeProperty("Object2")
            joint.removeProperty("Element2")
            joint.removeProperty("Vertex2")

            joint.addProperty(
                "App::PropertyXLinkSub",
                "Object2",
                "Joint Connector 2",
                QT_TRANSLATE_NOOP("App::Property", "The second object of the joint"),
                locked=True,
            )

            joint.Object2 = [obj2, [el2, vtx2]]

    def migrationScript2(self, joint):
        def processObject(object_attr, reference_attr, part_attr, connector_label, order):
            try:
                if hasattr(joint, object_attr):
                    joint.addProperty(
                        "App::PropertyXLinkSubHidden",
                        reference_attr,
                        connector_label,
                        QT_TRANSLATE_NOOP("App::Property", f"The {order} reference of the joint"),
                        locked=True,
                    )

                    obj = getattr(joint, object_attr)

                    base_obj = obj[0]
                    part = getattr(joint, part_attr)
                    elt = obj[1][0]
                    vtx = obj[1][1]

                    # Get the 'selection-root-obj' and the global path
                    root_obj, path = UtilsAssembly.getRootPath(base_obj, part)
                    base_obj = root_obj
                    elt = path + elt
                    vtx = path + vtx

                    setattr(joint, reference_attr, [base_obj, [elt, vtx]])

                    joint.removeProperty(object_attr)
                    joint.removeProperty(part_attr)

            except (AttributeError, IndexError, TypeError) as e:
                App.Console.PrintWarning(
                    f"{translate('Assembly', 'Assembly joint')} '{getContext(joint)}' "
                    f"{translate('Assembly', 'has an invalid')} '{object_attr}' "
                    f"{translate('Assembly', 'or related attributes')}. {str(e)}\n"
                )

        processObject("Object1", "Reference1", "Part1", "Joint Connector 1", "first")
        processObject("Object2", "Reference2", "Part2", "Joint Connector 2", "second")

    def migrationScript3(self, joint):
        if hasattr(joint, "Offset"):
            current_offset = joint.Offset  # App.Vector
            current_rotation = joint.Rotation  # float

            joint.removeProperty("Offset")
            joint.removeProperty("Rotation")

            joint.addProperty(
                "App::PropertyPlacement",
                "Offset1",
                "Joint Connector 1",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "This is the attachment offset of the first connector of the joint",
                ),
                locked=True,
            )

            joint.addProperty(
                "App::PropertyPlacement",
                "Offset2",
                "Joint Connector 2",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "This is the attachment offset of the second connector of the joint",
                ),
                locked=True,
            )

            joint.Offset2 = App.Placement(current_offset, App.Rotation(current_rotation, 0, 0))

    def migrationScript4(self, joint):
        def processReference(reference_attr):
            try:
                if hasattr(joint, reference_attr):
                    ref = getattr(joint, reference_attr)

                    doc_name = ref[0].Document.Name
                    sub1 = UtilsAssembly.fixBodyExtraFeatureInSub(doc_name, ref[1][0])
                    sub2 = UtilsAssembly.fixBodyExtraFeatureInSub(doc_name, ref[1][1])

                    if sub1 != ref[1][0] or sub2 != ref[1][1]:
                        setattr(joint, reference_attr, (ref[0], [sub1, sub2]))

            except (AttributeError, IndexError, TypeError) as e:
                App.Console.PrintWarning(
                    f"{translate('Assembly', 'Assembly joint')} '{getContext(joint)}' "
                    f"{translate('Assembly', 'has an invalid')} '{reference_attr}' "
                    f"{translate('Assembly', 'or related attributes')}. {str(e)}\n"
                )

        processReference("Reference1")
        processReference("Reference2")

    def migrationScript5(self, joint):
        if not joint.hasExtension("App::SuppressibleExtensionPython"):
            joint.addExtension("App::SuppressibleExtensionPython")

        if App.GuiUp:
            if not joint.ViewObject.hasExtension("Gui::ViewProviderSuppressibleExtensionPython"):
                joint.ViewObject.addExtension("Gui::ViewProviderSuppressibleExtensionPython")

        if hasattr(joint, "Activated"):
            activated = joint.Activated
            if not activated:
                print(
                    "The 'Activated' property has been replaced by the 'Suppressed' property. The file has a deactivated joint that is being migrated. If you open this file in an older version, it will not be deactivated anymore."
                )
            joint.removeProperty("Activated")
            joint.Suppressed = not activated

    def migrationScript6(self, joint):
        """
        This script migrates properties with fixed types from PropertyFloat to the
        correct PropertyLength or PropertyAngle type.
        """
        if (
            hasattr(joint, "Distance")
            and joint.getTypeIdOfProperty("Distance") == "App::PropertyFloat"
        ):
            old_value = joint.Distance
            joint.setPropertyStatus("Distance", "-LockDynamic")
            joint.removeProperty("Distance")
            self.addDistanceProperty(joint)

            # Additionally "Angle" joint is now using Angle property instead of Distance as it did
            if joint.JointType == "Angle":
                self.addAngleProperty(joint)
                joint.Angle = old_value
            else:
                joint.Distance = old_value

        if (
            hasattr(joint, "Distance2")
            and joint.getTypeIdOfProperty("Distance2") == "App::PropertyFloat"
        ):
            old_value = joint.Distance2
            joint.setPropertyStatus("Distance2", "-LockDynamic")
            joint.removeProperty("Distance2")
            self.addDistance2Property(joint)
            joint.Distance2 = old_value

        if (
            hasattr(joint, "LengthMin")
            and joint.getTypeIdOfProperty("LengthMin") == "App::PropertyFloat"
        ):
            old_value = joint.LengthMin
            joint.setPropertyStatus("LengthMin", "-LockDynamic")
            joint.removeProperty("LengthMin")
            self.addLengthMinProperty(joint)
            joint.LengthMin = old_value

        if (
            hasattr(joint, "LengthMax")
            and joint.getTypeIdOfProperty("LengthMax") == "App::PropertyFloat"
        ):
            old_value = joint.LengthMax
            joint.setPropertyStatus("LengthMax", "-LockDynamic")
            joint.removeProperty("LengthMax")
            self.addLengthMaxProperty(joint)
            joint.LengthMax = old_value

        if (
            hasattr(joint, "AngleMin")
            and joint.getTypeIdOfProperty("AngleMin") == "App::PropertyFloat"
        ):
            old_value = joint.AngleMin
            joint.setPropertyStatus("AngleMin", "-LockDynamic")
            joint.removeProperty("AngleMin")
            self.addAngleMinProperty(joint)
            joint.AngleMin = old_value

        if (
            hasattr(joint, "AngleMax")
            and joint.getTypeIdOfProperty("AngleMax") == "App::PropertyFloat"
        ):
            old_value = joint.AngleMax
            joint.setPropertyStatus("AngleMax", "-LockDynamic")
            joint.removeProperty("AngleMax")
            self.addAngleMaxProperty(joint)
            joint.AngleMax = old_value

    def migrationScript7(self, joint):
        """
        Migrates PropertyXLinkSubHidden (Assembly-rooted) to PropertyXLinkSub (Component-rooted).
        """

        def migrate_prop(prop_name, add_method, assembly):
            if not hasattr(joint, prop_name):
                return

            # Only migrate if it's the old type
            if joint.getTypeIdOfProperty(prop_name) != "App::PropertyXLinkSubHidden":
                return

            old_ref = getattr(joint, prop_name)

            joint.setPropertyStatus(prop_name, "-LockDynamic")
            joint.removeProperty(prop_name)

            add_method(joint)

            if old_ref:
                root = old_ref[0]
                old_subs = old_ref[1]
                if old_subs:
                    comp, first_new_sub = UtilsAssembly.getComponentReference(
                        assembly, root, old_subs[0]
                    )

                    if comp:
                        new_subs = [first_new_sub]

                        # We can deduce the prefix length from the difference
                        # between old and new of the first item.
                        prefix_len = len(old_subs[0]) - len(first_new_sub)

                        for i in range(1, len(old_subs)):
                            sub = old_subs[i]
                            if len(sub) >= prefix_len:
                                new_subs.append(sub[prefix_len:])
                            else:
                                new_subs.append(sub)

                        setattr(joint, prop_name, [comp, new_subs])

        assembly = self.getAssembly(joint)
        migrate_prop("Reference1", self.addReference1Property, assembly)
        migrate_prop("Reference2", self.addReference2Property, assembly)

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def getAssembly(self, joint):
        for obj in joint.InList:
            if obj.isDerivedFrom("Assembly::AssemblyObject"):
                return obj
            elif obj.isDerivedFrom("Assembly::AssemblyLink"):
                return self.getAssembly(obj)

        return None

    def setJointType(self, joint, newType):
        oldType = joint.JointType
        if newType != oldType:
            joint.JointType = newType

    def onChanged(self, joint, prop):
        """Do something when a property has changed"""
        # App.Console.PrintMessage("Change property: " + str(prop) + "\n")

        # during loading the onchanged may be triggered before full init.
        if App.isRestoring():
            return

        if prop == "JointType":
            newType = joint.JointType
            tr_new_type = TranslatedJointTypes[JointTypes.index(newType)]

            # Find the old joint type in the label and replace it
            for i, old_type_name in enumerate(JointTypes):
                if old_type_name == newType:
                    continue
                tr_old_type = TranslatedJointTypes[i]
                if tr_old_type in joint.Label:
                    joint.Label = joint.Label.replace(tr_old_type, tr_new_type)
                    break

        if prop == "Reference1" or prop == "Reference2":
            joint.recompute()

        if (
            not hasattr(joint, "Reference1")
            or not hasattr(joint, "Reference2")
            or joint.Reference1 is None
            or joint.Reference2 is None
        ):
            return

        if prop == "Offset1" or prop == "Offset2":
            self.updateJCSPlacements(joint)

            presolved = joint.JointType in JointUsingPreSolve and self.preSolve(joint, False)

            isAssembly = self.getAssembly(joint).Type == "Assembly"
            if isAssembly and not presolved:
                solveIfAllowed(self.getAssembly(joint))
            else:
                self.updateJCSPlacements(joint)

        if prop == "Distance" and joint.JointType == "Distance":
            solveIfAllowed(self.getAssembly(joint))

        if prop == "Angle" and joint.JointType == "Angle":
            if joint.Angle != 0.0:
                self.preventParallel(joint)
            solveIfAllowed(self.getAssembly(joint))

    def execute(self, joint):
        errStr = joint.Label + ": " + QT_TRANSLATE_NOOP("Assembly", "Broken link in: ")
        if (
            hasattr(joint, "Reference1")
            and joint.Reference1 is not None
            and len(joint.Reference1) == 2
            and len(joint.Reference1[1]) != 0
            and (joint.Reference1[1][0].find("?") != -1)
        ):
            raise Exception(errStr + "Reference1")

        if (
            hasattr(joint, "Reference2")
            and joint.Reference2 is not None
            and len(joint.Reference2) == 2
            and len(joint.Reference2[1]) != 0
            and (joint.Reference2[1][0].find("?") != -1)
        ):
            raise Exception(errStr + "Reference2")

        self.updateJCSPlacements(joint)

    def setJointConnectors(self, joint, refs):
        # current selection is a vector of strings like "Assembly.Assembly1.Assembly2.Body.Pad.Edge16" including both what selection return as obj_name and obj_sub
        assembly = self.getAssembly(joint)
        isAssembly = assembly.Type == "Assembly"

        if len(refs) >= 1:
            joint.Reference1 = refs[0]
            joint.Placement1 = self.findPlacement(joint, joint.Reference1, 0)
        else:
            joint.Reference1 = None
            joint.Placement1 = App.Placement()
            self.partMovedByPresolved = None

        if len(refs) >= 2:
            joint.Reference2 = refs[1]
            joint.Placement2 = self.findPlacement(joint, joint.Reference2, 1)
            if joint.JointType in JointUsingPreSolve:
                self.preSolve(joint)
            elif joint.JointType in JointParallelForbidden:
                self.preventParallel(joint)

            if isAssembly:
                solveIfAllowed(assembly, True)
            else:
                self.updateJCSPlacements(joint)

        else:
            joint.Reference2 = None
            joint.Placement2 = App.Placement()
            if isAssembly:
                assembly.undoSolve()
            self.undoPreSolve(joint)

    def updateJCSPlacements(self, joint):
        if not joint.Detach1:
            joint.Placement1 = self.findPlacement(joint, joint.Reference1, 0)

        if not joint.Detach2:
            joint.Placement2 = self.findPlacement(joint, joint.Reference2, 1)

        self.redrawJointPlacements(joint)

    def redrawJointPlacements(self, joint):
        if joint.ViewObject:
            proxy = joint.ViewObject.Proxy
            if proxy:
                proxy.redrawJointPlacements(joint)

    """
    So here we want to find a placement that corresponds to a local coordinate system that would be placed at the selected vertex.
    - obj is usually a App::Link to a PartDesign::Body, or primitive, fasteners. But can also be directly the object.1
    - elt can be a face, an edge or a vertex.
    - If elt is a vertex, then vtx = elt And placement is vtx coordinates without rotation.
    - if elt is an edge, then vtx = edge start/end vertex depending on which is closer. If elt is an arc or circle, vtx can also be the center. The rotation is the plane normal to the line positioned at vtx. Or for arcs/circle, the plane of the arc.
    - if elt is a plane face, vtx is the face vertex (to the list of vertex we need to add arc/circle centers) the closer to the mouse. The placement is the plane rotation positioned at vtx
    - if elt is a cylindrical face, vtx can also be the center of the arcs of the cylindrical face.
    """

    def findPlacement(self, joint, ref, index=0):
        ignoreVertex = joint.JointType == "Distance"
        plc = UtilsAssembly.findPlacement(ref, ignoreVertex)

        # We apply the attachment offsets.
        if index == 0:
            plc = plc * joint.Offset1
        else:
            plc = plc * joint.Offset2

        return plc

    def flipOnePart(self, joint):
        self.matchJCS(joint, False, True)

    def preSolve(self, joint, savePlc=True):
        # The goal of this is to put the part in the correct position to avoid wrong placement by the solve.

        # we actually don't want to match perfectly the JCS, it is best to match them
        # in the current closest direction, ie either matched or flipped.
        self.matchJCS(joint, savePlc)

    def matchJCS(self, joint, savePlc=True, reverse=False):
        assembly = self.getAssembly(joint)
        sameDir = self.areJcsSameDir(joint)
        if reverse:
            sameDir = not sameDir

        part1 = UtilsAssembly.getMovingPart(joint.Reference1)
        part2 = UtilsAssembly.getMovingPart(joint.Reference2)

        if not part1 or not part2:
            return False

        isAssembly = assembly.Type == "Assembly"
        if isAssembly:
            joint.Suppressed = True
            part1Connected = assembly.isPartConnected(part1)
            part2Connected = assembly.isPartConnected(part2)
            joint.Suppressed = False
        else:
            part1Connected = True
            part2Connected = False

        moving_part = None
        moving_part_ref = None
        fixed_part_ref = None
        moving_placement = None
        fixed_placement = None

        if not part1Connected and part1:
            moving_part = part1
            moving_part_ref = joint.Reference1
            fixed_part_ref = joint.Reference2
            moving_placement = joint.Placement1
            fixed_placement = joint.Placement2
        elif not part2Connected and part2:
            moving_part = part2
            moving_part_ref = joint.Reference2
            fixed_part_ref = joint.Reference1
            moving_placement = joint.Placement2
            fixed_placement = joint.Placement1
        else:
            # Both parts are constrained, or something is invalid. Nothing to pre-solve.
            return False

        parts_to_move = [moving_part]
        if isAssembly:
            parts_to_move = parts_to_move + assembly.getDownstreamParts(moving_part, joint)

        if savePlc:
            self.partsMovedByPresolved = {p: p.Placement for p in parts_to_move}

        moving_part_global_jcs = UtilsAssembly.getJcsGlobalPlc(moving_placement, moving_part_ref)
        fixed_part_global_jcs = UtilsAssembly.getJcsGlobalPlc(fixed_placement, fixed_part_ref)

        if not sameDir:
            moving_part_global_jcs = UtilsAssembly.flipPlacement(moving_part_global_jcs)

        transform_plc = fixed_part_global_jcs * moving_part_global_jcs.inverse()

        for part in parts_to_move:
            part.Placement = transform_plc * part.Placement

        return True

    def undoPreSolve(self, joint):
        if hasattr(self, "partsMovedByPresolved") and self.partsMovedByPresolved:
            for part, plc in self.partsMovedByPresolved.items():
                if part and hasattr(part, "Placement"):
                    part.Placement = plc
            self.partsMovedByPresolved = {}

            if joint.ViewObject:
                joint.ViewObject.Proxy.redrawJointPlacements(joint)

    def preventParallel(self, joint):
        # Angle and perpendicular joints in the solver cannot handle the situation where both JCS are Parallel
        parallel = self.areJcsZParallel(joint)
        if not parallel:
            return

        assembly = self.getAssembly(joint)

        part1 = UtilsAssembly.getMovingPart(joint.Reference1)
        part2 = UtilsAssembly.getMovingPart(joint.Reference2)

        isAssembly = assembly.Type == "Assembly"
        if isAssembly:
            part1ConnectedByJoint = assembly.isJointConnectingPartToGround(joint, "Reference1")
            part2ConnectedByJoint = assembly.isJointConnectingPartToGround(joint, "Reference2")
        else:
            part1ConnectedByJoint = False
            part2ConnectedByJoint = True

        if part2ConnectedByJoint:
            self.partMovedByPresolved = part2
            self.presolveBackupPlc = part2.Placement

            # Get the global JCS placement to find a suitable rotation axis (its own X-axis)
            globalJcsPlc = UtilsAssembly.getJcsGlobalPlc(joint.Placement2, joint.Reference2)
            # Transform the local X-axis vector (1,0,0) into the global coordinate system
            rotation_axis = globalJcsPlc.Rotation.multVec(App.Vector(1, 0, 0))

            part2.Placement = UtilsAssembly.applyRotationToPlacementAlongAxis(
                part2.Placement, 10, rotation_axis
            )

        elif part1ConnectedByJoint:
            self.partMovedByPresolved = part1
            self.presolveBackupPlc = part1.Placement

            # Get the global JCS placement to find a suitable rotation axis (its own X-axis)
            globalJcsPlc = UtilsAssembly.getJcsGlobalPlc(joint.Placement1, joint.Reference1)
            # Transform the local X-axis vector (1,0,0) into the global coordinate system
            rotation_axis = globalJcsPlc.Rotation.multVec(App.Vector(1, 0, 0))

            part1.Placement = UtilsAssembly.applyRotationToPlacementAlongAxis(
                part1.Placement, 10, rotation_axis
            )

    def areJcsSameDir(self, joint):
        globalJcsPlc1 = UtilsAssembly.getJcsGlobalPlc(joint.Placement1, joint.Reference1)
        globalJcsPlc2 = UtilsAssembly.getJcsGlobalPlc(joint.Placement2, joint.Reference2)

        return UtilsAssembly.arePlacementSameDir(globalJcsPlc1, globalJcsPlc2)

    def areJcsZParallel(self, joint):
        globalJcsPlc1 = UtilsAssembly.getJcsGlobalPlc(joint.Placement1, joint.Reference1)
        globalJcsPlc2 = UtilsAssembly.getJcsGlobalPlc(joint.Placement2, joint.Reference2)

        return UtilsAssembly.arePlacementZParallel(globalJcsPlc1, globalJcsPlc2)


class ViewProviderJoint:
    def __init__(self, vobj):
        """Set this object to the proxy object of the actual view provider"""

        vobj.Proxy = self

        vobj.addExtension("Gui::ViewProviderSuppressibleExtensionPython")

    def attach(self, vobj):
        """Setup the scene sub-graph of the view provider, this method is mandatory"""
        self.app_obj = vobj.Object

        self.switch_JCS1 = SoSwitchMarker(vobj)
        self.switch_JCS2 = SoSwitchMarker(vobj)
        self.switch_JCS_preview = SoSwitchMarker(vobj)

        self.display_mode = coin.SoType.fromName("SoFCSelection").createInstance()
        self.display_mode.addChild(self.switch_JCS1)
        self.display_mode.addChild(self.switch_JCS2)
        self.display_mode.addChild(self.switch_JCS_preview)
        vobj.addDisplayMode(self.display_mode, "Wireframe")

    def updateData(self, joint, prop):
        """If a property of the handled feature has changed we have the chance to handle this here"""
        if prop == "Placement1" and hasattr(joint, "Reference1"):
            self.redrawJointPlacement(self.switch_JCS1, joint.Placement1, joint.Reference1)

        if prop == "Placement2" and hasattr(joint, "Reference2"):
            self.redrawJointPlacement(self.switch_JCS2, joint.Placement2, joint.Reference2)

    def redrawJointPlacements(self, joint):
        if not hasattr(joint, "Reference1") or not hasattr(joint, "Reference2"):
            return

        self.redrawJointPlacement(self.switch_JCS1, joint.Placement1, joint.Reference1)
        self.redrawJointPlacement(self.switch_JCS2, joint.Placement2, joint.Reference2)

    def redrawJointPlacement(self, jcs, plc, ref):
        if ref:
            jcs.whichChild = coin.SO_SWITCH_ALL
            jcs.set_marker_placement(plc, ref)
        else:
            jcs.whichChild = coin.SO_SWITCH_NONE

    def showPreviewJCS(self, visible, placement=None, ref=None):
        if visible:
            self.switch_JCS_preview.whichChild = coin.SO_SWITCH_ALL
            self.switch_JCS_preview.set_marker_placement(placement, ref)
        else:
            self.switch_JCS_preview.whichChild = coin.SO_SWITCH_NONE

    def setPickableState(self, state: bool):
        """Set JCS selectable or unselectable in 3D view"""
        self.switch_JCS1.setPickableState(state)
        self.switch_JCS2.setPickableState(state)
        self.switch_JCS_preview.setPickableState(state)

    def getDisplayModes(self, obj):
        """Return a list of display modes."""
        modes = []
        modes.append("Wireframe")
        return modes

    def getDefaultDisplayMode(self):
        """Return the name of the default display mode. It must be defined in getDisplayModes."""
        return "Wireframe"

    def onChanged(self, vp, prop):
        """Here we can do something when a single property got changed"""
        # App.Console.PrintMessage("Change property: " + str(prop) + "\n")
        if prop == "color_X_axis" or prop == "color_Y_axis" or prop == "color_Z_axis":
            self.switch_JCS1.onChanged(vp, prop)
            self.switch_JCS2.onChanged(vp, prop)
            self.switch_JCS_preview.onChanged(vp, prop)

    def getIcon(self):
        if self.app_obj.JointType == "Fixed":
            return ":/icons/Assembly_CreateJointFixed.svg"
        elif self.app_obj.JointType == "Revolute":
            return ":/icons/Assembly_CreateJointRevolute.svg"
        elif self.app_obj.JointType == "Cylindrical":
            return ":/icons/Assembly_CreateJointCylindrical.svg"
        elif self.app_obj.JointType == "Slider":
            return ":/icons/Assembly_CreateJointSlider.svg"
        elif self.app_obj.JointType == "Ball":
            return ":/icons/Assembly_CreateJointBall.svg"
        elif self.app_obj.JointType == "Distance":
            return ":/icons/Assembly_CreateJointDistance.svg"
        elif self.app_obj.JointType == "Parallel":
            return ":/icons/Assembly_CreateJointParallel.svg"
        elif self.app_obj.JointType == "Perpendicular":
            return ":/icons/Assembly_CreateJointPerpendicular.svg"
        elif self.app_obj.JointType == "Angle":
            return ":/icons/Assembly_CreateJointAngle.svg"
        elif self.app_obj.JointType == "RackPinion":
            return ":/icons/Assembly_CreateJointRackPinion.svg"
        elif self.app_obj.JointType == "Screw":
            return ":/icons/Assembly_CreateJointScrew.svg"
        elif self.app_obj.JointType == "Gears":
            return ":/icons/Assembly_CreateJointGears.svg"
        elif self.app_obj.JointType == "Belt":
            return ":/icons/Assembly_CreateJointPulleys.svg"

        return ":/icons/Assembly_CreateJoint.svg"

    def getOverlayIcons(self):
        """
        Return a dictionary of overlay icons.
        Keys are positions from Gui.IconPosition.
        Values are the icon resource names.
        """

        overlays = {}

        assembly = self.app_obj.Proxy.getAssembly(self.app_obj)
        # Assuming Reference1 corresponds to the first part link
        if hasattr(self.app_obj, "Reference1"):
            part = UtilsAssembly.getMovingPart(self.app_obj.Reference1)
            if part is not None and not assembly.isPartConnected(part):
                overlays[Gui.IconPosition.BottomLeft] = "Part_Detached"

        return overlays

    def dumps(self):
        """When saving the document this object gets stored using Python's json module.\
                Since we have some un-serializable parts here -- the Coin stuff -- we must define this method\
                to return a tuple of all serializable objects or None."""
        return None

    def loads(self, state):
        """When restoring the serialized object from document we have the chance to set some internals here.\
                Since no data were serialized nothing needs to be done here."""
        return None

    def doubleClicked(self, vobj):
        App.closeActiveTransaction(True)  # Close the auto-transaction

        task = Gui.Control.activeTaskDialog()
        if task:
            task.reject()

        assembly = vobj.Object.Proxy.getAssembly(vobj.Object)

        if assembly is None:
            return False

        if UtilsAssembly.activeAssembly() != assembly:
            vobj.Document.setEdit(assembly)

        panel = TaskAssemblyCreateJoint(0, vobj.Object)
        dialog = Gui.Control.showDialog(panel)
        if dialog is not None:
            dialog.setAutoCloseOnTransactionChange(True)
            dialog.setAutoCloseOnDeletedDocument(True)
            dialog.setDocumentName(App.ActiveDocument.Name)

        return True

    def canDelete(self, _obj):
        return True


################ Grounded Joint object #################


class GroundedJoint:
    def __init__(self, joint, obj_to_ground):
        joint.Proxy = self
        self.joint = joint

        self.createObjectToGroundProperty(joint, obj_to_ground)

    def createObjectToGroundProperty(self, joint, obj_to_ground):
        joint.addProperty(
            "App::PropertyLinkGlobal",
            "ObjectToGround",
            "Ground",
            QT_TRANSLATE_NOOP("App::Property", "The object to ground"),
            locked=True,
        )
        joint.ObjectToGround = obj_to_ground

    def onDocumentRestored(self, joint):
        self.migrationScript(joint)

    def migrationScript(self, joint):
        if (
            hasattr(joint, "ObjectToGround")
            and joint.getTypeIdOfProperty("ObjectToGround") == "App::PropertyLink"
        ):
            obj_to_ground = joint.ObjectToGround
            joint.setPropertyStatus("ObjectToGround", "-LockDynamic")
            joint.removeProperty("ObjectToGround")

            self.createObjectToGroundProperty(joint, obj_to_ground)

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def onChanged(self, fp, prop):
        """Do something when a property has changed"""
        # App.Console.PrintMessage("Change property: " + str(prop) + "\n")
        pass

    def execute(self, fp):
        """Do something when doing a recomputation, this method is mandatory"""
        # App.Console.PrintMessage("Recompute Python Box feature\n")
        pass


class ViewProviderGroundedJoint:
    def __init__(self, obj):
        """Set this object to the proxy object of the actual view provider"""
        obj.Proxy = self

    def attach(self, vobj):
        """Setup the scene sub-graph of the view provider, this method is mandatory"""
        app_obj = vobj.Object
        if app_obj is None:
            return
        groundedObj = app_obj.ObjectToGround
        if groundedObj is None:
            return

        self.scaleFactor = 3.0

        lockpadColorInt = Preferences.preferences().GetUnsigned("AssemblyConstraints", 0xCC333300)
        self.lockpadColor = coin.SoBaseColor()
        self.lockpadColor.rgb.setValue(UtilsAssembly.color_from_unsigned(lockpadColorInt))

        self.app_obj = vobj.Object
        app_doc = self.app_obj.Document
        self.gui_doc = Gui.getDocument(app_doc)

        # Create transformation (position and orientation)
        self.transform = coin.SoTransform()
        self.set_lock_position(groundedObj)

        # Create the 2D components of the lockpad: a square and two arcs
        self.square = self.create_square()

        # Creating the arcs (approximated with line segments)
        self.arc = self.create_arc(0, 4, 4, 0, 180)

        self.pick = coin.SoPickStyle()
        self.pick.style.setValue(coin.SoPickStyle.SHAPE_ON_TOP)

        # Assemble the parts into a scenegraph
        self.lockpadSeparator = coin.SoSeparator()
        self.lockpadSeparator.addChild(self.lockpadColor)
        self.lockpadSeparator.addChild(self.square)
        self.lockpadSeparator.addChild(self.arc)

        # Use SoVRMLBillboard to make sure the lockpad always faces the camera
        self.billboard = coin.SoVRMLBillboard()
        self.billboard.addChild(self.lockpadSeparator)

        self.scale = coin.SoType.fromName("SoShapeScale").createInstance()
        self.scale.setPart("shape", self.billboard)
        self.scale.scaleFactor = self.scaleFactor

        self.transformSeparator = coin.SoSeparator()
        self.transformSeparator.addChild(self.transform)
        self.transformSeparator.addChild(self.pick)
        self.transformSeparator.addChild(self.scale)

        # Attach the scenegraph to the view provider
        vobj.addDisplayMode(self.transformSeparator, "Wireframe")

    def create_square(self):
        coords = [
            (-5, -4, 0),
            (5, -4, 0),
            (5, 4, 0),
            (-5, 4, 0),
        ]
        vertices = coin.SoCoordinate3()
        vertices.point.setValues(0, 4, coords)

        squareFace = coin.SoFaceSet()
        squareFace.numVertices.setValue(4)

        square = coin.SoAnnotation()
        square.addChild(vertices)
        square.addChild(squareFace)

        return square

    def create_arc(self, centerX, centerY, radius, startAngle, endAngle):
        coords = []
        for angle in range(
            startAngle, endAngle + 1, 5
        ):  # Increment can be adjusted for smoother arcs
            rad = math.radians(angle)
            x = centerX + math.cos(rad) * radius
            y = centerY + math.sin(rad) * radius
            coords.append((x, y, 0))

        radius = radius * 0.7
        for angle in range(endAngle + 1, startAngle - 1, -5):  # Step backward
            rad = math.radians(angle)
            x = centerX + math.cos(rad) * radius
            y = centerY + math.sin(rad) * radius
            coords.append((x, y, 0))

        vertices = coin.SoCoordinate3()
        vertices.point.setValues(0, len(coords), coords)

        shapeHints = coin.SoShapeHints()
        shapeHints.faceType = coin.SoShapeHints.UNKNOWN_FACE_TYPE

        line = coin.SoFaceSet()
        line.numVertices.setValue(len(coords))

        arc = coin.SoAnnotation()
        arc.addChild(shapeHints)
        arc.addChild(vertices)
        arc.addChild(line)

        return arc

    def set_lock_position(self, groundedObj):
        bBox = groundedObj.ViewObject.getBoundingBox()
        if bBox.isValid():
            pos = bBox.Center
        else:
            pos = groundedObj.Placement.Base

        self.transform.translation.setValue(pos.x, pos.y, pos.z)

    def updateData(self, fp, prop):
        """If a property of the handled feature has changed we have the chance to handle this here"""
        # fp is the handled feature, prop is the name of the property that has changed

        if prop == "Placement" and fp.ObjectToGround:
            self.set_lock_position(fp.ObjectToGround)

    def getDisplayModes(self, obj):
        """Return a list of display modes."""
        modes = ["Wireframe"]
        return modes

    def getDefaultDisplayMode(self):
        """Return the name of the default display mode. It must be defined in getDisplayModes."""
        return "Wireframe"

    def onChanged(self, vp, prop):
        """Here we can do something when a single property got changed"""
        # App.Console.PrintMessage("Change property: " + str(prop) + "\n")
        pass

    def getIcon(self):
        return ":/icons/Assembly_ToggleGrounded.svg"

    def dumps(self):
        """When saving the document this object gets stored using Python's json module.\
                Since we have some un-serializable parts here -- the Coin stuff -- we must define this method\
                to return a tuple of all serializable objects or None."""
        return None

    def loads(self, state):
        """When restoring the serialized object from document we have the chance to set some internals here.\
                Since no data were serialized nothing needs to be done here."""
        return None

    def canDelete(self, _obj):
        return True


class MakeJointSelGate:
    def __init__(self, taskbox, assembly):
        self.taskbox = taskbox
        self.assembly = assembly

    def allow(self, doc, obj, sub):
        if not sub:
            return False

        objs_names, element_name = UtilsAssembly.getObjsNamesAndElement(obj.Name, sub)

        if self.assembly.Name not in objs_names:
            # Only objects within the assembly.
            return False

        ref = [obj, [sub]]
        sel_obj = UtilsAssembly.getObject(ref)

        if UtilsAssembly.isLink(sel_obj):
            linked = sel_obj.getLinkedObject()
            if linked == sel_obj:
                return True  # We accept empty links
            sel_obj = linked

        if sel_obj.isDerivedFrom("Part::Feature") or sel_obj.isDerivedFrom("App::Part"):
            return True

        if sel_obj.isDerivedFrom("App::LocalCoordinateSystem") or sel_obj.isDerivedFrom(
            "App::DatumElement"
        ):
            datum = sel_obj
            if datum.isDerivedFrom("App::DatumElement"):
                parent = datum.getParent()
                if parent.isDerivedFrom("App::LocalCoordinateSystem"):
                    datum = parent

            if self.assembly.hasObject(datum) and hasattr(datum, "MapMode"):
                # accept only datum that are not attached
                return datum.MapMode == "Deactivated"

            return True

        return False


activeTask = None


class TaskAssemblyCreateJoint(QtCore.QObject):
    def __init__(self, jointTypeIndex, jointObj=None, subclass=False):
        super().__init__()

        global activeTask
        activeTask = self
        self.blockOffsetRotation = False

        self.assembly = UtilsAssembly.activeAssembly()
        if not self.assembly:
            self.assembly = UtilsAssembly.activePart()
            self.activeType = "Part"
        else:
            self.activeType = "Assembly"
            self.assembly.ensureIdentityPlacements()

        self.doc = self.assembly.Document
        self.gui_doc = Gui.getDocument(self.doc)

        self.view = self.gui_doc.activeView()

        if not self.assembly or not self.view or not self.doc:
            return

        if self.activeType == "Assembly":
            self.assembly.ViewObject.MoveOnlyPreselected = True
            self.assembly.ViewObject.MoveInCommand = False

        # Create a top-level container widget for subclasses of TaskAssemblyCreateJoint
        self.form = QtWidgets.QWidget()

        # Load the joint creation UI and parent it to `self.form`
        self.jForm = Gui.PySideUic.loadUi(":/panels/TaskAssemblyCreateJoint.ui", self.form)

        # Create a layout for `self.form` and add `self.jForm` to it
        layout = QtWidgets.QVBoxLayout(self.form)
        if not subclass:
            layout.setContentsMargins(0, 0, 0, 0)
            layout.setSpacing(0)
        layout.addWidget(self.jForm)

        self.isolate_modes = ["Transparent", "Wireframe", "Hidden", "Disabled"]
        self.jForm.isolateType.addItems(
            [translate("Assembly", mode) for mode in self.isolate_modes]
        )
        self.jForm.isolateType.currentIndexChanged.connect(self.updateIsolation)

        if self.activeType == "Part":
            self.jForm.setWindowTitle("Match parts")
            self.jForm.jointType.hide()
            self.jForm.isolateType.hide()

        self.jForm.jointType.addItems(TranslatedJointTypes)

        self.jForm.jointType.setCurrentIndex(jointTypeIndex)
        self.jType = JointTypes[self.jForm.jointType.currentIndex()]
        self.jForm.jointType.currentIndexChanged.connect(self.onJointTypeChanged)

        if jointObj:
            Gui.Selection.clearSelection()
            self.creating = False
            self.joint = jointObj
            self.jointName = jointObj.Label
            App.setActiveTransaction("Edit " + self.jointName + " Joint")

            self.updateTaskboxFromJoint()
            self.visibilityBackup = self.joint.Visibility
            self.joint.Visibility = True

        else:
            self.creating = True
            self.jointName = self.jForm.jointType.currentText().replace(" ", "")
            if self.activeType == "Part":
                App.setActiveTransaction("Transform")
            else:
                App.setActiveTransaction("Create " + self.jointName + " Joint")

            self.refs = []
            self.presel_ref = None

            self.createJointObject()
            self.visibilityBackup = False

        self.jForm.angleSpinbox.valueChanged.connect(self.onAngleChanged)
        self.jForm.distanceSpinbox.valueChanged.connect(self.onDistanceChanged)
        self.jForm.distanceSpinbox2.valueChanged.connect(self.onDistance2Changed)
        self.jForm.offsetSpinbox.valueChanged.connect(self.onOffsetChanged)
        self.jForm.rotationSpinbox.valueChanged.connect(self.onRotationChanged)
        bind = Gui.ExpressionBinding(self.jForm.angleSpinbox).bind(self.joint, "Angle")
        bind = Gui.ExpressionBinding(self.jForm.distanceSpinbox).bind(self.joint, "Distance")
        bind = Gui.ExpressionBinding(self.jForm.distanceSpinbox2).bind(self.joint, "Distance2")
        bind = Gui.ExpressionBinding(self.jForm.offsetSpinbox).bind(self.joint, "Offset2.Base.z")
        bind = Gui.ExpressionBinding(self.jForm.rotationSpinbox).bind(
            self.joint, "Offset2.Rotation.Yaw"
        )

        self.jForm.reverseRotCheckbox.setChecked(self.jType == "Gears")
        self.jForm.reverseRotCheckbox.stateChanged.connect(self.reverseRotToggled)

        self.jForm.advancedOffsetCheckbox.stateChanged.connect(self.advancedOffsetToggled)

        self.jForm.offset1Button.clicked.connect(self.onOffset1Clicked)
        self.jForm.offset2Button.clicked.connect(self.onOffset2Clicked)
        self.jForm.PushButtonReverse.clicked.connect(self.onReverseClicked)

        self.jForm.limitCheckbox1.stateChanged.connect(self.adaptUi)
        self.jForm.limitCheckbox2.stateChanged.connect(self.adaptUi)
        self.jForm.limitCheckbox3.stateChanged.connect(self.adaptUi)
        self.jForm.limitCheckbox4.stateChanged.connect(self.adaptUi)

        self.jForm.limitLenMinSpinbox.valueChanged.connect(self.onLimitLenMinChanged)
        self.jForm.limitLenMaxSpinbox.valueChanged.connect(self.onLimitLenMaxChanged)
        self.jForm.limitRotMinSpinbox.valueChanged.connect(self.onLimitRotMinChanged)
        self.jForm.limitRotMaxSpinbox.valueChanged.connect(self.onLimitRotMaxChanged)
        bind = Gui.ExpressionBinding(self.jForm.limitLenMinSpinbox).bind(self.joint, "LengthMin")
        bind = Gui.ExpressionBinding(self.jForm.limitLenMaxSpinbox).bind(self.joint, "LengthMax")
        bind = Gui.ExpressionBinding(self.jForm.limitRotMinSpinbox).bind(self.joint, "AngleMin")
        bind = Gui.ExpressionBinding(self.jForm.limitRotMaxSpinbox).bind(self.joint, "AngleMax")

        self.adaptUi()

        if self.creating:
            # This has to be after adaptUi so that properties default values are adapted
            # if needed. For instance for gears adaptUi will prevent radii from being 0
            # before handleInitialSelection tries to solve.
            self.handleInitialSelection()

        UtilsAssembly.setJointsPickableState(self.doc, False)

        Gui.Selection.addSelectionGate(
            MakeJointSelGate(self, self.assembly), Gui.Selection.ResolveMode.NoResolve
        )
        Gui.Selection.addObserver(self, Gui.Selection.ResolveMode.NoResolve)
        Gui.Selection.setSelectionStyle(Gui.Selection.SelectionStyle.GreedySelection)

        self.callbackMove = self.view.addEventCallback("SoLocation2Event", self.moveMouse)
        self.callbackKey = self.view.addEventCallback("SoKeyboardEvent", self.KeyboardEvent)

        self.jForm.featureList.installEventFilter(self)

        self.createDeleteAction()

        self.addition_rejected = False

    def accept(self):
        if len(self.refs) != 2:
            App.Console.PrintWarning(
                translate("Assembly", "Select 2 elements from 2 separate parts")
            )
            return False

        self.deactivate()

        if self.activeType == "Assembly":
            self.joint.Visibility = self.visibilityBackup
        else:
            self.joint.Document.removeObject(self.joint.Name)

        cmds = UtilsAssembly.generatePropertySettings(self.joint)
        Gui.doCommand(cmds)

        self.assembly.recompute(True)

        App.closeActiveTransaction()
        return True

    def reject(self):
        self.deactivate()
        App.closeActiveTransaction(True)
        self.assembly.recompute(True)
        return True

    def autoClosedOnTransactionChange(self):
        self.reject()

    def autoClosedOnDeletedDocument(self):
        global activeTask
        activeTask = None
        Gui.Selection.removeSelectionGate()
        Gui.Selection.removeObserver(self)
        Gui.Selection.setSelectionStyle(Gui.Selection.SelectionStyle.NormalSelection)
        App.closeActiveTransaction(True)

    def deactivate(self):
        global activeTask
        activeTask = None

        if self.activeType == "Assembly":
            self.assembly.clearUndo()
            self.assembly.ViewObject.MoveOnlyPreselected = False
            self.assembly.ViewObject.MoveInCommand = True

        Gui.Selection.removeSelectionGate()
        Gui.Selection.removeObserver(self)
        Gui.Selection.setSelectionStyle(Gui.Selection.SelectionStyle.NormalSelection)
        Gui.Selection.clearSelection()
        self.view.removeEventCallback("SoLocation2Event", self.callbackMove)
        self.view.removeEventCallback("SoKeyboardEvent", self.callbackKey)
        UtilsAssembly.setJointsPickableState(self.doc, True)
        if Gui.Control.activeDialog():
            Gui.Control.closeDialog()

    def handleInitialSelection(self):
        selection = Gui.Selection.getSelectionEx("*", 0)
        if not selection:
            return
        for sel in selection:
            # If you select 2 solids (bodies for example) within an assembly.
            # There'll be a single sel but 2 SubElementNames.

            if not sel.SubElementNames:
                # no subnames, so its a root assembly itself that is selected.
                Gui.Selection.removeSelection(sel.Object)
                continue

            for sub_name in sel.SubElementNames:
                # We add sub_name twice because the joints references have element name + vertex name
                # and in the case of initial selection, both are the same.

                moving_part, new_sub = UtilsAssembly.getComponentReference(
                    self.assembly, sel.Object, sub_name
                )
                if not moving_part:
                    break

                # Construct the reference using the Component as the root
                ref = [moving_part, [new_sub, new_sub]]

                # Only objects within the assembly.
                if moving_part is None:
                    Gui.Selection.removeSelection(sel.Object, sub_name)
                    continue

                if len(self.refs) == 1 and moving_part == self.getMovingPart(self.refs[0]):
                    # do not select several feature of the same object.
                    self.refs.clear()
                    Gui.Selection.clearSelection()
                    return

                self.refs.append(ref)

        # do not accept initial selection if we don't have 2 selected features
        if len(self.refs) != 2:
            self.refs.clear()
            Gui.Selection.clearSelection()
        else:
            self.updateJoint()

    def createJointObject(self):
        type_index = self.jForm.jointType.currentIndex()

        if self.activeType == "Part":
            self.joint = self.assembly.newObject("App::FeaturePython", "Temporary joint")
        else:
            joint_group = UtilsAssembly.getJointGroup(self.assembly)
            self.joint = joint_group.newObject("App::FeaturePython", "Joint")
            self.joint.Label = self.jointName
            joint_group.purgeTouched()
            self.assembly.purgeTouched()

        Joint(self.joint, type_index)
        ViewProviderJoint(self.joint.ViewObject)
        self.joint.purgeTouched()

    def onJointTypeChanged(self, index):
        self.jType = JointTypes[self.jForm.jointType.currentIndex()]
        self.joint.Proxy.setJointType(self.joint, self.jType)
        self.adaptUi()

    def onAngleChanged(self, quantity):
        self.joint.Angle = self.jForm.angleSpinbox.property("rawValue")

    def onDistanceChanged(self, quantity):
        self.joint.Distance = self.jForm.distanceSpinbox.property("rawValue")

    def onDistance2Changed(self, quantity):
        self.joint.Distance2 = self.jForm.distanceSpinbox2.property("rawValue")

    def onOffsetChanged(self, quantity):
        if self.blockOffsetRotation:
            return

        self.joint.Offset2.Base = App.Vector(0, 0, self.jForm.offsetSpinbox.property("rawValue"))

    def onRotationChanged(self, quantity):
        if self.blockOffsetRotation:
            return

        yaw = self.jForm.rotationSpinbox.property("rawValue")
        ypr = self.joint.Offset2.Rotation.getYawPitchRoll()
        self.joint.Offset2.Rotation.setYawPitchRoll(yaw, ypr[1], ypr[2])

    def onLimitLenMinChanged(self, quantity):
        if self.jForm.limitCheckbox1.isChecked():
            self.joint.LengthMin = self.jForm.limitLenMinSpinbox.property("rawValue")

    def onLimitLenMaxChanged(self, quantity):
        if self.jForm.limitCheckbox2.isChecked():
            self.joint.LengthMax = self.jForm.limitLenMaxSpinbox.property("rawValue")

    def onLimitRotMinChanged(self, quantity):
        if self.jForm.limitCheckbox3.isChecked():
            self.joint.AngleMin = self.jForm.limitRotMinSpinbox.property("rawValue")

    def onLimitRotMaxChanged(self, quantity):
        if self.jForm.limitCheckbox4.isChecked():
            self.joint.AngleMax = self.jForm.limitRotMaxSpinbox.property("rawValue")

    def onReverseClicked(self):
        self.joint.Proxy.flipOnePart(self.joint)

    def reverseRotToggled(self, val):
        if val:
            self.jForm.jointType.setCurrentIndex(JointTypes.index("Gears"))
        else:
            self.jForm.jointType.setCurrentIndex(JointTypes.index("Belt"))

    def adaptUi(self):
        jType = self.jType

        needAngle = jType in JointUsingAngle
        self.jForm.angleLabel.setVisible(needAngle)
        self.jForm.angleSpinbox.setVisible(needAngle)

        needDistance = jType in JointUsingDistance
        self.jForm.distanceLabel.setVisible(needDistance)
        self.jForm.distanceSpinbox.setVisible(needDistance)
        if needDistance:
            if jType == "Distance":
                self.jForm.distanceLabel.setText(translate("Assembly", "Distance"))
            elif jType == "Gears" or jType == "Belt":
                self.jForm.distanceLabel.setText(translate("Assembly", "Radius 1"))
            elif jType == "Screw":
                self.jForm.distanceLabel.setText(translate("Assembly", "Thread pitch"))
            else:
                self.jForm.distanceLabel.setText(translate("Assembly", "Pitch radius"))

        needDistance2 = jType in JointUsingDistance2
        self.jForm.distanceLabel2.setVisible(needDistance2)
        self.jForm.distanceSpinbox2.setVisible(needDistance2)
        self.jForm.reverseRotCheckbox.setVisible(needDistance2)

        if jType in JointNoNegativeDistance:
            # Setting minimum to 0.01 to prevent 0 and negative values
            self.jForm.distanceSpinbox.setProperty("minimum", 1e-7)
            if self.jForm.distanceSpinbox.property("rawValue") == 0.0:
                self.jForm.distanceSpinbox.setProperty("rawValue", 1.0)

            if jType == "Gears" or jType == "Belt":
                self.jForm.distanceSpinbox2.setProperty("minimum", 1e-7)
                if self.jForm.distanceSpinbox2.property("rawValue") == 0.0:
                    self.jForm.distanceSpinbox2.setProperty("rawValue", 1.0)
        else:
            self.jForm.distanceSpinbox.setProperty("minimum", float("-inf"))
            self.jForm.distanceSpinbox2.setProperty("minimum", float("-inf"))

        advancedOffset = self.jForm.advancedOffsetCheckbox.isChecked()
        needOffset = jType in JointUsingOffset
        needRotation = jType in JointUsingRotation
        self.jForm.offset1Label.setVisible(advancedOffset)
        self.jForm.offset2Label.setVisible(advancedOffset)
        self.jForm.offset1Button.setVisible(advancedOffset)
        self.jForm.offset2Button.setVisible(advancedOffset)
        self.jForm.offsetLabel.setVisible(not advancedOffset and needOffset)
        self.jForm.offsetSpinbox.setVisible(not advancedOffset and needOffset)
        self.jForm.rotationLabel.setVisible(not advancedOffset and needRotation)
        self.jForm.rotationSpinbox.setVisible(not advancedOffset and needRotation)

        self.jForm.PushButtonReverse.setVisible(jType in JointUsingReverse)

        needLengthLimits = jType in JointUsingLimitLength
        needAngleLimits = jType in JointUsingLimitAngle
        needLimits = needLengthLimits or needAngleLimits
        self.jForm.groupBox_limits.setVisible(needLimits)

        if needLimits:
            self.joint.EnableLengthMin = self.jForm.limitCheckbox1.isChecked()
            self.joint.EnableLengthMax = self.jForm.limitCheckbox2.isChecked()
            self.joint.EnableAngleMin = self.jForm.limitCheckbox3.isChecked()
            self.joint.EnableAngleMax = self.jForm.limitCheckbox4.isChecked()

            self.jForm.limitCheckbox1.setVisible(needLengthLimits)
            self.jForm.limitCheckbox2.setVisible(needLengthLimits)
            self.jForm.limitLenMinSpinbox.setVisible(needLengthLimits)
            self.jForm.limitLenMaxSpinbox.setVisible(needLengthLimits)

            self.jForm.limitCheckbox3.setVisible(needAngleLimits)
            self.jForm.limitCheckbox4.setVisible(needAngleLimits)
            self.jForm.limitRotMinSpinbox.setVisible(needAngleLimits)
            self.jForm.limitRotMaxSpinbox.setVisible(needAngleLimits)

            if needLengthLimits:
                self.jForm.limitLenMinSpinbox.setEnabled(self.joint.EnableLengthMin)
                self.jForm.limitLenMaxSpinbox.setEnabled(self.joint.EnableLengthMax)
                self.onLimitLenMinChanged(0)  # dummy value
                self.onLimitLenMaxChanged(0)

            if needAngleLimits:
                self.jForm.limitRotMinSpinbox.setEnabled(self.joint.EnableAngleMin)
                self.jForm.limitRotMaxSpinbox.setEnabled(self.joint.EnableAngleMax)
                self.onLimitRotMinChanged(0)
                self.onLimitRotMaxChanged(0)

        self.updateOffsetWidgets()

    def updateOffsetWidgets(self):
        # Makes sure the values in both the simplified and advanced tabs are sync.
        pos = self.joint.Offset1.Base
        self.jForm.offset1Button.setText(f"({pos.x}, {pos.y}, {pos.z})")

        pos = self.joint.Offset2.Base
        self.jForm.offset2Button.setText(f"({pos.x}, {pos.y}, {pos.z})")

        self.blockOffsetRotation = True
        self.jForm.offsetSpinbox.setProperty("rawValue", pos.z)
        self.jForm.rotationSpinbox.setProperty(
            "rawValue", self.joint.Offset2.Rotation.getYawPitchRoll()[0]
        )
        self.blockOffsetRotation = False

    def advancedOffsetToggled(self, on):
        self.adaptUi()
        self.updateOffsetWidgets()

    def onOffset1Clicked(self):
        UtilsAssembly.openEditingPlacementDialog(self.joint, "Offset1")
        self.updateOffsetWidgets()

    def onOffset2Clicked(self):
        UtilsAssembly.openEditingPlacementDialog(self.joint, "Offset2")
        self.updateOffsetWidgets()

    def updateIsolation(self):
        """Isolates the two selected components or clears isolation."""

        if self.activeType != "Assembly":
            return

        isolate_mode = self.jForm.isolateType.currentIndex()

        assembly_vobj = self.assembly.ViewObject

        # If "Disabled" is selected, clear any active isolation and stop.
        if isolate_mode == 3:
            assembly_vobj.clearIsolate()
            return

        if len(self.refs) == 2:
            try:
                # Use a set to handle cases where both refs point to the same object
                parts_to_isolate = {
                    self.getMovingPart(self.refs[0]),
                    self.getMovingPart(self.refs[1]),
                }
                assembly_vobj.isolateComponents(list(parts_to_isolate), isolate_mode)
            except Exception as e:
                App.Console.PrintWarning(f"Could not update isolation: {e}\n")
                assembly_vobj.clearIsolate()
        else:
            assembly_vobj.clearIsolate()

    def updateTaskboxFromJoint(self):
        self.refs = []
        self.presel_ref = None

        ref1 = self.joint.Reference1
        ref2 = self.joint.Reference2

        if UtilsAssembly.isRefValid(ref1, 2):
            self.refs.append(ref1)
            sub1 = UtilsAssembly.addTipNameToSub(ref1)
            Gui.Selection.addSelection(ref1[0].Document.Name, ref1[0].Name, sub1)

        if UtilsAssembly.isRefValid(ref2, 2):
            self.refs.append(ref2)
            sub2 = UtilsAssembly.addTipNameToSub(ref2)
            Gui.Selection.addSelection(ref2[0].Document.Name, ref2[0].Name, sub2)

        self.jForm.angleSpinbox.setProperty("rawValue", self.joint.Angle.Value)
        self.jForm.distanceSpinbox.setProperty("rawValue", self.joint.Distance.Value)
        self.jForm.distanceSpinbox2.setProperty("rawValue", self.joint.Distance2.Value)
        self.jForm.offsetSpinbox.setProperty("rawValue", self.joint.Offset2.Base.z)
        self.jForm.rotationSpinbox.setProperty(
            "rawValue", self.joint.Offset2.Rotation.getYawPitchRoll()[0]
        )

        self.jForm.limitCheckbox1.setChecked(self.joint.EnableLengthMin)
        self.jForm.limitCheckbox2.setChecked(self.joint.EnableLengthMax)
        self.jForm.limitCheckbox3.setChecked(self.joint.EnableAngleMin)
        self.jForm.limitCheckbox4.setChecked(self.joint.EnableAngleMax)
        self.jForm.limitLenMinSpinbox.setProperty("rawValue", self.joint.LengthMin.Value)
        self.jForm.limitLenMaxSpinbox.setProperty("rawValue", self.joint.LengthMax.Value)
        self.jForm.limitRotMinSpinbox.setProperty("rawValue", self.joint.AngleMin.Value)
        self.jForm.limitRotMaxSpinbox.setProperty("rawValue", self.joint.AngleMax.Value)

        self.jForm.jointType.setCurrentIndex(JointTypes.index(self.joint.JointType))
        self.updateJointList()
        self.updateIsolation()

    def updateJoint(self):
        # First we build the listwidget
        self.updateJointList()

        # Then we pass the new list to the joint object
        self.joint.Proxy.setJointConnectors(self.joint, self.refs)

        self.updateIsolation()

    def updateJointList(self):
        self.jForm.featureList.clear()
        simplified_names = []
        for ref in self.refs:

            sname = UtilsAssembly.getObject(ref).Label

            element_name = UtilsAssembly.getElementName(ref[1][0])
            if element_name != "":
                sname = sname + "." + element_name
            simplified_names.append(sname)
        self.jForm.featureList.addItems(simplified_names)

    def updateLimits(self):
        needLengthLimits = self.jType in JointUsingLimitLength
        needAngleLimits = self.jType in JointUsingLimitAngle
        if needLengthLimits:
            distance = UtilsAssembly.getJointDistance(self.joint)
            if (
                not self.jForm.limitCheckbox1.isChecked()
                and self.jForm.limitLenMinSpinbox.property("expression") == ""
            ):
                self.jForm.limitLenMinSpinbox.setProperty("rawValue", distance)
            if (
                not self.jForm.limitCheckbox2.isChecked()
                and self.jForm.limitLenMaxSpinbox.property("expression") == ""
            ):
                self.jForm.limitLenMaxSpinbox.setProperty("rawValue", distance)

        if needAngleLimits:
            angle = UtilsAssembly.getJointXYAngle(self.joint) / math.pi * 180
            if (
                not self.jForm.limitCheckbox3.isChecked()
                and self.jForm.limitRotMinSpinbox.property("expression") == ""
            ):
                self.jForm.limitRotMinSpinbox.setProperty("rawValue", angle)
            if (
                not self.jForm.limitCheckbox4.isChecked()
                and self.jForm.limitRotMaxSpinbox.property("expression") == ""
            ):
                self.jForm.limitRotMaxSpinbox.setProperty("rawValue", angle)

    def moveMouse(self, info):
        if len(self.refs) >= 2 or (
            len(self.refs) == 1
            and (
                not self.presel_ref
                or self.getMovingPart(self.refs[0]) == self.getMovingPart(self.presel_ref)
            )
        ):
            self.joint.ViewObject.Proxy.showPreviewJCS(False)
            if len(self.refs) >= 2:
                self.updateLimits()
            return

        cursor_pos = self.view.getCursorPos()
        cursor_info = self.view.getObjectInfo(cursor_pos)
        # cursor_info example  {'x': 41.515, 'y': 7.449, 'z': 16.861, 'ParentObject': <Part object>, 'SubName': 'Body002.Pad.Face5', 'Document': 'part3', 'Object': 'Pad', 'Component': 'Face5'}

        if (
            not cursor_info
            or not self.presel_ref
            # or cursor_info["SubName"] != self.presel_ref["sub_name"]
            # Removed because they are not equal when hovering a line endpoints.
            # But we don't actually need to test because if there's no preselection then not cursor is None
        ):
            self.joint.ViewObject.Proxy.showPreviewJCS(False)
            return

        ref = self.presel_ref

        # newPos = self.view.getPoint(*info["Position"]) is not OK: it's not pos on the object but on the focal plane
        newPos = App.Vector(cursor_info["x"], cursor_info["y"], cursor_info["z"])
        vertex_name = UtilsAssembly.findElementClosestVertex(ref, newPos)

        ref = UtilsAssembly.addVertexToReference(ref, vertex_name)

        placement = self.joint.Proxy.findPlacement(self.joint, ref, 0)
        self.joint.ViewObject.Proxy.showPreviewJCS(True, placement, ref)
        self.previewJCSVisible = True

    # 3D view keyboard handler
    def KeyboardEvent(self, info):
        if info["State"] == "UP" and info["Key"] == "ESCAPE":
            self.reject()

        if info["State"] == "UP" and info["Key"] == "RETURN":
            self.accept()

    def _removeSelectedItems(self, selected_indexes):
        for index in selected_indexes:
            row = index.row()
            if row < len(self.refs):
                ref = self.refs[row]
                sub = UtilsAssembly.addTipNameToSub(ref)
                Gui.Selection.removeSelection(ref[0], sub)
            else:
                print(f"Row {row} is out of bounds for refs (length: {len(self.refs)})")

    def eventFilter(self, watched, event):
        if self.jForm is not None and watched == self.jForm.featureList:
            if event.type() == QtCore.QEvent.ShortcutOverride:
                if (
                    hasattr(self, "deleteAction")
                    and self.deleteAction.shortcut().matches(event.key())
                    != QtGui.QKeySequence.NoMatch
                ):
                    event.accept()
                    return True
                return False

            elif event.type() == QtCore.QEvent.KeyPress:
                if (
                    hasattr(self, "deleteAction")
                    and self.deleteAction.shortcut().matches(event.key())
                    != QtGui.QKeySequence.NoMatch
                ):
                    self.deleteAction.trigger()
                    return True  # Consume the event

        return super().eventFilter(watched, event)

    def createDeleteAction(self):
        """Create delete action with shortcut"""
        try:
            delete_sequence = Gui.QtTools.deleteKeySequence()
        except AttributeError:
            # fallback to standard key if there is no sequence defined
            delete_sequence = QtGui.QKeySequence(QtCore.Qt.Key_Delete)

        self.deleteAction = QtGui.QAction("Remove", self.jForm)
        self.deleteAction.setShortcut(delete_sequence)

        self.deleteAction.setIcon(
            QtWidgets.QApplication.style().standardIcon(QtWidgets.QStyle.SP_DialogCancelButton)
        )

        self.deleteAction.setShortcutVisibleInContextMenu(True)

        self.jForm.featureList.addAction(self.deleteAction)
        self.jForm.featureList.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)

        self.deleteAction.triggered.connect(self.deleteSelectedItems)

    def deleteSelectedItems(self):
        """Delete selected items from the feature list - same logic as Delete key in ev filter"""
        selected_indexes = self.jForm.featureList.selectedIndexes()
        self._removeSelectedItems(selected_indexes)

    def getMovingPart(self, ref):
        return UtilsAssembly.getMovingPart(ref)

    # selectionObserver stuff
    def addSelection(self, doc_name, obj_name, sub_name, mousePos):
        rootObj = App.getDocument(doc_name).getObject(obj_name)

        # We do not need the full TNP string like :"Part.Body.Pad.;#a:1;:G0;XTR;:Hc94:8,F.Face6"
        # instead we need : "Part.Body.Pad.Face6"
        resolved = rootObj.resolveSubElement(sub_name, True)
        sub_name = resolved[2]

        sub_name = UtilsAssembly.fixBodyExtraFeatureInSub(doc_name, sub_name)

        comp, new_sub = UtilsAssembly.getComponentReference(self.assembly, rootObj, sub_name)
        if not comp:
            # Selection was not valid (not inside assembly or logic failed)
            Gui.Selection.removeSelection(doc_name, obj_name, sub_name)
            return

        # Construct the reference using the Component as the root
        ref = [comp, [new_sub]]

        moving_part = self.getMovingPart(ref)

        # Check if the addition is acceptable (we are not doing this in selection gate to let user move objects)
        acceptable = True
        if len(self.refs) >= 2:
            # No more than 2 elements can be selected for basic joints.
            acceptable = False

        for reference in self.refs:
            sel_moving_part = self.getMovingPart(reference)
            if sel_moving_part == moving_part:
                # Can't join a solid to itself. So the user need to select 2 different parts.
                acceptable = False

        if not acceptable:
            self.addition_rejected = True
            Gui.Selection.removeSelection(doc_name, obj_name, sub_name)
            return

        # Selection is acceptable so add it

        mousePos = App.Vector(mousePos[0], mousePos[1], mousePos[2])
        vertex_name = UtilsAssembly.findElementClosestVertex(ref, mousePos)

        # add the vertex name to the reference
        ref = UtilsAssembly.addVertexToReference(ref, vertex_name)

        self.refs.append(ref)
        self.updateJoint()

        # We hide the preview JCS if we just added to the selection
        self.joint.ViewObject.Proxy.showPreviewJCS(False)

    def removeSelection(self, doc_name, obj_name, sub_name, mousePos=None):
        if self.addition_rejected:
            self.addition_rejected = False
            return

        rootObj = App.getDocument(doc_name).getObject(obj_name)

        # Apply the same processing as in addSelection to ensure consistent comparison
        resolved = rootObj.resolveSubElement(sub_name, True)
        sub_name = resolved[2]

        sub_name = UtilsAssembly.fixBodyExtraFeatureInSub(doc_name, sub_name)

        comp, new_sub = UtilsAssembly.getComponentReference(self.assembly, rootObj, sub_name)
        if not comp:
            return

        for reference in self.refs[:]:
            ref_obj = reference[0]
            ref_element_name = reference[1][0] if len(reference[1]) > 0 else ""

            # match both object and processed element name for precise identification
            if ref_obj == comp and ref_element_name == new_sub:
                self.refs.remove(reference)
                break
        else:
            print("No matching ref found for removal!")

        self.updateJoint()

    def setPreselection(self, doc_name, obj_name, sub_name):
        if not sub_name:
            self.presel_ref = None
            return

        rootObj = App.getDocument(doc_name).getObject(obj_name)

        comp, new_sub = UtilsAssembly.getComponentReference(self.assembly, rootObj, sub_name)
        if not comp:
            return

        self.presel_ref = [comp, [new_sub]]

    def clearSelection(self, doc_name):
        self.refs.clear()
        self.updateJoint()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
from PySide import QtCore

from PySide.QtCore import QT_TRANSLATE_NOOP

__doc__ = """CAM Copy object and FreeCAD command"""


translate = FreeCAD.Qt.translate


class ObjectPathCopy:
    def __init__(self, obj):
        obj.addProperty(
            "App::PropertyLink",
            "Base",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "The toolpath to be copied"),
        )
        obj.addProperty(
            "App::PropertyLink",
            "ToolController",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "The tool controller that will be used to calculate the toolpath",
            ),
        )
        obj.Proxy = self

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def execute(self, obj):
        if obj.Base:
            if hasattr(obj.Base, "ToolController"):
                obj.ToolController = obj.Base.ToolController
            if obj.Base.Path:
                obj.Path = obj.Base.Path.copy()
            if obj.Base.Placement:
                obj.Placement = obj.Base.Placement


class ViewProviderPathCopy:
    def __init__(self, vobj):
        self.Object = vobj.Object
        vobj.Proxy = self

    def attach(self, vobj):
        self.Object = vobj.Object
        return

    def getIcon(self):
        return ":/icons/CAM_Copy.svg"

    def dumps(self):
        return None

    def loads(self, state):
        return None


class CommandPathCopy:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Copy",
            "MenuText": QT_TRANSLATE_NOOP("CAM_Copy", "Copy"),
            "ToolTip": QT_TRANSLATE_NOOP("CAM_Copy", "Creates a linked copy of another toolpath"),
        }

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False

    def Activated(self):

        FreeCAD.ActiveDocument.openTransaction("Create Copy")
        FreeCADGui.addModule("Path.Op.Gui.Copy")

        consolecode = """
import Path
import Path.Op.Gui.Copy
selGood = True
# check that the selection contains exactly what we want
selection = FreeCADGui.Selection.getSelection()
proj = selection[0].InList[0] #get the group that the selectied object is inside

if len(selection) != 1:
    FreeCAD.Console.PrintError(translate("CAM_Copy", "Please select one toolpath object")+"\n")
    selGood = False

if not selection[0].isDerivedFrom("Path::Feature"):
    FreeCAD.Console.PrintError(translate("CAM_Copy", "The selected object is not a toolpath")+"\n")
    selGood = False

if selGood:
    obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", str(selection[0].Name)+'_Copy')
    Path.Op.Gui.Copy.ObjectPathCopy(obj)
    Path.Op.Gui.Copy.ViewProviderPathCopy(obj.ViewObject)
    obj.Base = FreeCAD.ActiveDocument.getObject(selection[0].Name)
    if hasattr(obj.Base, 'ToolController'):
        obj.ToolController = obj.Base.ToolController

g = proj.Group
g.append(obj)
proj.Group = g

FreeCAD.ActiveDocument.recompute()

"""

        FreeCADGui.doCommand(consolecode)
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_Copy", CommandPathCopy())

FreeCAD.Console.PrintLog("Loading PathCopy… done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# -*- coding: utf8 -*-
# ***************************************************************************
# *   Copyright (c) 2009 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2018 George Shuklin (amarao)                            *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to return the SVG representation of text elements."""
## @package svgtext
# \ingroup draftfunctions
# \brief Provides functions to return the SVG representation of text elements.

import math

import FreeCAD as App
import DraftVecUtils
import draftutils.utils as utils

## \addtogroup draftfunctions
# @{


def _get_text_techdraw(text, tcolor, fontsize, anchor, align, fontname, angle, base, linespacing):
    """Return the SVG representation of text for TechDraw display.

    `text` is a list of textual elements; they are iterated, styled,
    and added around a `<text>` tag.
    ::
        <text ...> text[0] </text>
        <text ...> text[1] </text>
    """
    svg = ""
    for i in range(len(text)):
        delta = App.Vector(0, -i * linespacing, 0)
        point = base + DraftVecUtils.rotate(delta, angle)

        _t = text[i].replace("&", "&amp;")
        _t = _t.replace("<", "&lt;")
        t = _t.replace(">", "&gt;")

        svg += "<text "
        svg += 'stroke-width="0" stroke="{}" '.format(tcolor)
        svg += 'fill="{}" font-size="{}" '.format(tcolor, fontsize)
        svg += 'style="text-anchor:{};text-align:{};'.format(anchor, align.lower())
        svg += 'font-family:{}" '.format(fontname)
        svg += 'transform="'
        svg += "rotate({},{},{}) ".format(math.degrees(angle), point.x, point.y)
        svg += "translate({},{}) ".format(point.x, point.y)
        svg += 'scale(1,-1)"'
        # svg += 'freecad:skip="1"'
        svg += ">\n"
        svg += t
        svg += "</text>\n"
    return svg


def _get_text_header(tcolor, fontsize, anchor, align, fontname, angle, base, flip):
    """Return the initial <text> tag with style options.

    The text must be added after this tag, and then must be closed.
    ::
        <text ...>
        ...
        </text>
    """
    svg = "<text "
    svg += 'stroke-width="0" stroke="{}" '.format(tcolor)
    svg += 'fill="{}" font-size="{}" '.format(tcolor, fontsize)
    svg += 'style="text-anchor:{};text-align:{};'.format(anchor, align.lower())
    svg += 'font-family:{}" '.format(fontname)
    svg += 'transform="'
    svg += "rotate({},{},{}) ".format(math.degrees(angle), base.x, base.y)
    if flip:
        svg += "translate({},{}) ".format(base.x, base.y)
    else:
        svg += "translate({},{}) ".format(base.x, -base.y)
    # svg += 'scale({},-{}) '.format(tmod/2000, tmod/2000)

    if flip:
        svg += "scale(1,-1) "
    else:
        svg += "scale(1,1) "

    svg += '" '
    svg += 'freecad:skip="1"'
    svg += ">\n"
    return svg


def get_text(
    plane,
    techdraw,
    tcolor,
    fontsize,
    fontname,
    angle,
    base,
    text,
    linespacing=0.5,
    align="center",
    flip=True,
):
    """Get the SVG representation of a textual element."""
    if isinstance(angle, App.Rotation):
        if not plane:
            angle = angle.Angle
        else:
            if plane.axis.getAngle(angle.Axis) < 0.001:
                angle = angle.Angle
            elif abs(plane.axis.getAngle(angle.Axis) - math.pi) < 0.001:
                if abs(angle.Angle) > 0.1:
                    angle = -angle.Angle
                else:
                    angle = angle.Angle
            elif abs(plane.axis.getAngle(angle.Axis) - math.pi / 2) < 0.001:
                # text is perpendicular to view, so it shouldn't appear
                return ""
            else:
                # Compute an X vector
                vec = App.Vector(1, 0, 0)
                vec = angle.multVec(App.Vector(1, 0, 0))
                angle = DraftVecUtils.angle(vec, plane.u)

    # text should be a list of strings separated by a newline
    if not isinstance(text, list):
        text = text.split("\n")

    if align.lower() == "center":
        anchor = "middle"
    elif align.lower() == "left":
        anchor = "start"
    else:
        anchor = "end"

    if techdraw:
        # For TechDraw display each item in the text list is placed
        # in an individual tag.
        # <text ...> text[0] </text>
        # <text ...> text[1] </text>
        svg = _get_text_techdraw(
            text, tcolor, fontsize, anchor, align, fontname, angle, base, linespacing
        )
    else:
        # If the SVG is not for TechDraw, and there is a single item
        # in the text list, place it in a single tag.
        # <text ...> text </text>
        #
        # For multiple elements, place each element inside a <tspan> tag.
        # <text ...>
        #   <tspan>text[0]</tspan>
        #   <tspan>text[1]</tspan>
        # </text>
        svg = _get_text_header(tcolor, fontsize, anchor, align, fontname, angle, base, flip)

        if len(text) == 1:
            _t = text[0].replace("&", "&amp;").replace("<", "&lt;")
            svg += _t.replace(">", "&gt;")
        else:
            for i in range(len(text)):
                if i == 0:
                    svg += "<tspan>"
                else:
                    svg += '<tspan x="0" dy="{}">'.format(linespacing)
                _t = text[i].replace("&", "&amp;").replace("<", "&lt;")
                svg += _t.replace(">", "&gt;")
                svg += "</tspan>\n"
        svg += "</text>\n"
    return svg


def getText(
    plane,
    techdraw,
    tcolor,
    fontsize,
    fontname,
    angle,
    base,
    text,
    linespacing=0.5,
    align="center",
    flip=True,
):
    """Get the SVG representation of a textual element. DEPRECATED."""
    utils.use_instead("get_text")
    return get_text(
        plane, techdraw, tcolor, fontsize, fontname, angle, base, text, linespacing, align, flip
    )


## @}



# ***************************************************************************
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Result import for Z88 displacements format"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package importZ88O2Results
#  \ingroup FEM
#  \brief FreeCAD Z88 Disp Reader for FEM workbench

import os

import FreeCAD
from FreeCAD import Console
from builtins import open as pyopen


# ********* generic FreeCAD import and export methods *********


def open(filename):
    "called when freecad opens a file"
    docname = os.path.splitext(os.path.basename(filename))[0]
    insert(filename, docname)


def insert(filename, docname):
    "called when freecad wants to import a file"
    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    FreeCAD.ActiveDocument = doc
    import_z88_disp(filename)


# ********* module specific methods *********
def import_z88_disp(filename, analysis=None, result_name_prefix=None):
    """insert a FreeCAD FEM mechanical result object in the ActiveDocument
    pure usage:
    import feminout.importZ88O2Results as importZ88O2Results
    disp_file = "/pathtofile/z88o2.txt"
    importZ88O2Results.import_z88_disp(disp_file)

    the z888i1.txt FEMMesh file needs to be in the same directory as z88o2.txt
    # ahh, make a new document first ;-)
    """
    import ObjectsFem
    from . import importZ88Mesh
    from . import importToolsFem
    from femresult import resulttools

    if result_name_prefix is None:
        result_name_prefix = ""
    disp_read = read_z88_disp(filename)
    result_mesh_object = None
    if len(disp_read["Nodes"]) > 0:
        if analysis:
            analysis_object = analysis

        # read result mesh
        if filename.endswith("z88o2.txt"):
            mesh_file = filename.replace("o2", "i1")
            mesh_data = importZ88Mesh.read_z88_mesh(mesh_file)
            femmesh = importToolsFem.make_femmesh(mesh_data)
            result_mesh_object = ObjectsFem.makeMeshResult(FreeCAD.ActiveDocument, "Result_mesh")
            result_mesh_object.FemMesh = femmesh
        else:
            Console.PrintError("Z88 mesh file z88i1.txt not found.\n")
            return None

        # create result obj
        for result_set in disp_read["Results"]:
            results_name = result_name_prefix + "results"

            res_obj = ObjectsFem.makeResultMechanical(FreeCAD.ActiveDocument, results_name)
            res_obj.Mesh = result_mesh_object
            res_obj = importToolsFem.fill_femresult_mechanical(res_obj, result_set)
            res_obj = resulttools.add_disp_apps(res_obj)  # fill DisplacementLengths
            res_obj = resulttools.fill_femresult_stats(res_obj)  # fill Stats

            if analysis:
                analysis_object.addObject(res_obj)

        if FreeCAD.GuiUp:
            if analysis:
                import FemGui

                FemGui.setActiveAnalysis(analysis_object)
            FreeCAD.ActiveDocument.recompute()

    else:
        Console.PrintError(
            "Problem on Z88 result file import. No nodes found in Z88 result file.\n"
        )
        return None
    return res_obj


def read_z88_disp(z88_disp_input):
    """
    read a z88 disp file and extract the nodes and elements
    z88 Displacement output file is z88o2.txt
    works with Z88OS14
    """
    nodes = {}
    mode_disp = {}
    mode_results = {}
    results = []

    z88_disp_file = pyopen(z88_disp_input, "r")

    for no, line in enumerate(z88_disp_file):
        lno = no + 1
        linelist = line.split()

        if lno >= 6:
            # disp line
            node_no = int(linelist[0])
            mode_disp_x = float(linelist[1])
            mode_disp_y = float(linelist[2])
            if len(linelist) > 3:
                mode_disp_z = float(linelist[3])
            else:
                mode_disp_z = 0.0
            mode_disp[node_no] = FreeCAD.Vector(mode_disp_x, mode_disp_y, mode_disp_z)
            nodes[node_no] = node_no

    mode_results["disp"] = mode_disp
    results.append(mode_results)

    for r in results[0]["disp"]:
        Console.PrintLog("{} --> {}\n".format(r, results[0]["disp"][r]))

    z88_disp_file.close()
    return {"Nodes": nodes, "Results": results}



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint tie"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


from FreeCAD import Units, Vector


def get_analysis_types():
    return "all"  # write for all analysis types


def get_sets_name():
    return "constraints_tie_surface_sets"


def get_constraint_title():
    return "Tie Constraints"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return ""


def get_after_write_constraint():
    return ""


def write_meshdata_constraint(f, femobj, tie_obj, ccxwriter):
    # slave DEP
    f.write(f"*SURFACE, NAME=TIE_DEP{tie_obj.Name}\n")
    for refs, surf, is_sub_el in femobj["TieSlaveFaces"]:
        if is_sub_el:
            for elem, fno in surf:
                f.write(f"{elem},S{fno}\n")
        else:
            for elem in surf:
                f.write(f"{elem},S2\n")

    # master IND
    f.write(f"*SURFACE, NAME=TIE_IND{tie_obj.Name}\n")
    for refs, surf, is_sub_el in femobj["TieMasterFaces"]:
        if is_sub_el:
            for elem, fno in surf:
                f.write(f"{elem},S{fno}\n")
        else:
            for elem in surf:
                f.write(f"{elem},S2\n")


def write_constraint(f, femobj, tie_obj, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module

    tolerance = tie_obj.Tolerance.getValueAs("mm").Value
    adjust = ""
    symmetry = ""
    tie_name = tie_obj.Name
    if not tie_obj.Adjust:
        adjust = ", ADJUST=NO"

    if tie_obj.CyclicSymmetry:
        symmetry = ", CYCLIC SYMMETRY"

    f.write(
        "*TIE, POSITION TOLERANCE={:.13G}{}{}, NAME=TIE{}\n".format(
            tolerance, adjust, symmetry, tie_name
        )
    )
    ind_surf = f"TIE_IND{tie_name}"
    dep_surf = f"TIE_DEP{tie_name}"
    f.write(f"{dep_surf}, {ind_surf}\n")

    # write CYCLIC SYMMETRY MODEL card
    if tie_obj.CyclicSymmetry:
        f.write(
            "*CYCLIC SYMMETRY MODEL, N={}, NGRAPH={}, TIE=TIE{}, ELSET=Eall\n".format(
                tie_obj.Sectors, tie_obj.ConnectedSectors, tie_name
            )
        )

        # get symmetry axis points
        vec_a = tie_obj.SymmetryAxis.Base
        vec_b = tie_obj.SymmetryAxis * Vector(0, 0, 1)

        set_unit = lambda x: Units.Quantity(x, Units.Length).getValueAs("mm").Value
        point_a = [set_unit(coord) for coord in vec_a]
        point_b = [set_unit(coord) for coord in vec_b]

        f.write("{:.13G},{:.13G},{:.13G},{:.13G},{:.13G},{:.13G}\n".format(*point_a, *point_b))



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD

import Fem

from .truss_3d_cs_circle_ele_seg3 import setup as setup_truss_seg3
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Truss 3D circle cs, seg2 elements",
        "meshtype": "edge",
        "meshelement": "Seg3",
        "constraints": ["fixed", "force"],
        "solvers": ["z88"],
        "material": "solid",
        "equations": ["mechanical"],
    }


def setup(doc=None, solvertype="z88"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # see seg3 for the header
    # for simplicity reason only one explanation is used for both

    # setup cantilever faceload
    doc = setup_truss_seg3(doc, solvertype)
    femmesh_obj = doc.getObject(get_meshname())

    # mesh
    from .meshes.mesh_truss_crane_seg2 import create_nodes, create_elements

    fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)

    # overwrite mesh with the hexa20 mesh
    femmesh_obj.FemMesh = fem_mesh

    # set changed properties
    femmesh_obj.ElementOrder = "1st"
    # one element for each bar
    femmesh_obj.CharacteristicLengthMax = "3000.0 mm"
    femmesh_obj.CharacteristicLengthMin = "3000.0 mm"

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

from .camotics import CamoticsToolBitSerializer
from .fctb import FCTBSerializer
from .yaml import YamlToolBitSerializer


all_serializers = (
    CamoticsToolBitSerializer,
    FCTBSerializer,
    YamlToolBitSerializer,
)


__all__ = [
    "CamoticsToolBitSerializer",
    "FCTBSerializer",
    "YamlToolBitSerializer",
    "all_serializers",
]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2023 David Carter <dcarter@david.carter.ca>             *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# **************************************************************************/

__title__ = "material model utilities"
__author__ = "David Carter"
__url__ = "https://www.freecad.org"

import os
import io
from pathlib import Path
import yaml

import FreeCAD


unicode = str

__models = {}
__modelsByPath = {}

def _dereference(parent, child):
    # Add the child parameters to the parent
    parentModel = parent["model"]
    parentBase = parent["base"]
    childModel = child["model"]
    childBase = child["base"]
    for name, value in childModel[childBase].items():
        if name not in ["Name", "UUID", "URL", "Description", "DOI", "Inherits"] and \
            name not in parentModel[parentBase]: # Don't add if it's already there
            parentModel[parentBase][name] = value

    print("dereferenced:")
    print(parentModel)

def _dereferenceInheritance(data):
    if not data["dereferenced"]:
        data["dereferenced"] = True # Prevent recursion loops

        model = data["model"]
        base = data["base"]
        if "Inherits" in model[base]:
            print("Model '{0}' inherits from:".format(data["name"]))
            for parent in model[base]["Inherits"]:
                print("\t'{0}'".format(parent))
                print("\t\t'{0}'".format(parent.keys()))
                print("\t\t'{0}'".format(parent["UUID"]))

                # This requires that all models have already been loaded undereferenced
                child = __models[parent["UUID"]]
                if child is not None:
                    _dereference(data, child)

def _dereferenceAll():
    for data in __models.values():
        _dereferenceInheritance(data)

def _scanFolder(folder):
    print("Scanning folder '{0}'".format(folder.absolute()))
    for child in folder.iterdir():
        if child.is_dir():
            _scanFolder(child)
        else:
            if child.suffix.lower() == ".yml":
                data = getModelFromPath(child)

                if data is not None:
                    __models[data["uuid"]] = data
                    __modelsByPath[data["path"]] = data
                    # print(data["model"])
            else:
                print("Extension '{0}'".format(child.suffix.lower()))

def _scanModels(libraries):
    __models = {} # Clear the current library
    __modelsByPath = {}
    print("_scanModels")
    print(libraries)
    for library in libraries:
        _scanFolder(Path(library))

    # Satisfy aany inheritances
    _dereferenceAll()

def getPreferredSaveDirectory():
    pass

def getModelLibraries():

    libraries = []

    # TODO: Expand beyond the standard models as we do for material paths
    path = Path(FreeCAD.getResourceDir()) / "Mod/Material/Resources/Models"
    libraries.append(path)

    _scanModels(libraries)

    return libraries

def getModel(uuid):
    """
        Retrieve the specified model.
    """
    if len(__models) < 1:
        getModelLibraries()

    if uuid not in __models:
        return None
    return __models[uuid]

def getModelFromPath(filePath):
    """
        Retrieve the model at the specified path.

        This may not need public exposure?
    """
    try:
        path = Path(filePath)
        stream = open(path.absolute(), "r")
        model = yaml.safe_load(stream)

        base = "Model"
        if "AppearanceModel" in model:
            base = "AppearanceModel"

        uuid = model[base]["UUID"]
        name = model[base]["Name"]

        data = {}
        data["base"] = base
        data["name"] = name
        data["path"] = path.absolute()
        data["uuid"] = uuid
        data["model"] = model
        data["dereferenced"] = False
        return data
    except Exception as ex:
        print("Unable to load '{0}'".format(path.absolute()))
        print(ex)

    return None

def saveModel(model, path):
    """
        Write the model to the specified path
    """


# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest
import json
import FreeCAD
from CAMTests.PathTestUtils import PathTestWithAssets
from Path.Tool.library import Library
from Path.Tool.toolbit import ToolBitEndmill
from Path.Tool.shape import ToolBitShapeEndmill, ToolBitShapeBallend
from Path.Tool.library.serializers import CamoticsLibrarySerializer, LinuxCNCSerializer


class TestPathToolLibrarySerializerBase(PathTestWithAssets):
    """Base class for Library serializer tests."""

    def setUp(self):
        super().setUp()
        self.test_library_id = "test_library"
        self.test_library_label = "Test Library"
        self.test_library = Library(self.test_library_label, id=self.test_library_id)

        # Create some dummy tool bits
        shape1 = ToolBitShapeEndmill("endmill_1")
        shape1.set_parameter("Diameter", FreeCAD.Units.Quantity("6.0 mm"))
        shape1.set_parameter("Length", FreeCAD.Units.Quantity("20.0 mm"))
        tool1 = ToolBitEndmill(shape1, id="tool_1")
        tool1.label = "Endmill 6mm"

        shape2 = ToolBitShapeEndmill("endmill_2")
        shape2.set_parameter("Diameter", FreeCAD.Units.Quantity("3.0 mm"))
        shape2.set_parameter("Length", FreeCAD.Units.Quantity("15.0 mm"))
        tool2 = ToolBitEndmill(shape2, id="tool_2")
        tool2.label = "Endmill 3mm"

        shape3 = ToolBitShapeBallend("ballend_1")
        shape3.set_parameter("Diameter", FreeCAD.Units.Quantity("5.0 mm"))
        shape3.set_parameter("Length", FreeCAD.Units.Quantity("18.0 mm"))
        tool3 = ToolBitEndmill(shape3, id="tool_3")
        tool3.label = "Ballend 5mm"

        self.test_library.add_bit(tool1, 1)
        self.test_library.add_bit(tool2, 2)
        self.test_library.add_bit(tool3, 3)


class TestCamoticsLibrarySerializer(TestPathToolLibrarySerializerBase):
    """Tests for the CamoticsLibrarySerializer."""

    def test_camotics_serialize(self):
        serializer = CamoticsLibrarySerializer
        serialized_data = serializer.serialize(self.test_library)
        self.assertIsInstance(serialized_data, bytes)

        # Verify the content structure (basic check)
        data_dict = json.loads(serialized_data.decode("utf-8"))
        self.assertEqual(
            data_dict["1"],
            {
                "description": "Endmill 6mm",
                "diameter": 6.0,
                "length": 20.0,
                "shape": "Cylindrical",
                "units": "metric",
            },
        )
        self.assertEqual(
            data_dict["2"],
            {
                "description": "Endmill 3mm",
                "diameter": 3.0,
                "length": 15.0,
                "shape": "Cylindrical",
                "units": "metric",
            },
        )
        self.assertEqual(
            data_dict["3"],
            {
                "description": "Ballend 5mm",
                "diameter": 5.0,
                "length": 18.0,
                "shape": "Ballnose",
                "units": "metric",
            },
        )

    def test_camotics_deserialize(self):
        serializer = CamoticsLibrarySerializer
        # Create a dummy serialized data matching the expected format
        dummy_data = {
            "10": {
                "units": "metric",
                "shape": "Ballnose",
                "length": 25,
                "diameter": 8,
                "description": "Ballnose 8mm",
            },
            "20": {
                "units": "metric",
                "shape": "Cylindrical",
                "length": 30,
                "diameter": 10,
                "description": "Endmill 10mm",
            },
        }
        dummy_bytes = json.dumps(dummy_data, indent=2).encode("utf-8")

        # Deserialize the data
        deserialized_library = serializer.deserialize(dummy_bytes, "deserialized_lib", {})

        self.assertIsInstance(deserialized_library, Library)
        self.assertEqual(deserialized_library.get_id(), "deserialized_lib")
        self.assertEqual(len(deserialized_library._bit_nos), 2)

        tool_10 = deserialized_library._bit_nos.get(10)
        assert tool_10 is not None, "tool not in the library"
        self.assertEqual(tool_10.label, "Ballnose 8mm")
        self.assertEqual(tool_10._tool_bit_shape.name, "Ballend")
        self.assertEqual(
            tool_10._tool_bit_shape.get_parameter("Diameter"), FreeCAD.Units.Quantity("8 mm")
        )
        self.assertEqual(
            tool_10._tool_bit_shape.get_parameter("Length"), FreeCAD.Units.Quantity("25 mm")
        )

        tool_20 = deserialized_library._bit_nos.get(20)
        assert tool_20 is not None, "tool not in the library"
        self.assertEqual(tool_20.label, "Endmill 10mm")
        self.assertEqual(tool_20._tool_bit_shape.name, "Endmill")
        self.assertEqual(
            tool_20._tool_bit_shape.get_parameter("Diameter"), FreeCAD.Units.Quantity("10 mm")
        )
        self.assertEqual(
            tool_20._tool_bit_shape.get_parameter("Length"), FreeCAD.Units.Quantity("30 mm")
        )


class TestLinuxCNCLibrarySerializer(TestPathToolLibrarySerializerBase):
    """Tests for the LinuxCNCLibrarySerializer."""

    def test_linuxcnc_serialize(self):
        serializer = LinuxCNCSerializer
        serialized_data = serializer.serialize(self.test_library)
        self.assertIsInstance(serialized_data, bytes)

        # Verify the content format (basic check)
        lines = serialized_data.decode("ascii", "ignore").strip().split("\n")
        self.assertEqual(len(lines), 3)
        self.assertEqual(lines[0], "T1 P0 X0 Y0 Z0 A0 B0 C0 U0 V0 W0 D6.00 I0 J0 Q0 ;Endmill 6mm")
        self.assertEqual(lines[1], "T2 P0 X0 Y0 Z0 A0 B0 C0 U0 V0 W0 D3.00 I0 J0 Q0 ;Endmill 3mm")
        self.assertEqual(lines[2], "T3 P0 X0 Y0 Z0 A0 B0 C0 U0 V0 W0 D5.00 I0 J0 Q0 ;Ballend 5mm")

    def test_linuxcnc_deserialize_not_implemented(self):
        serializer = LinuxCNCSerializer
        dummy_data = b"T1 P0 X0 Y0 Z0 A0 B0 C0 U0 V0 W0 D6.00 I0 J0 Q0 ;Endmill 6mm\n"
        with self.assertRaises(NotImplementedError):
            serializer.deserialize(dummy_data, "dummy_id", {})


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the viewprovider code for the polar Array object.

Currently unused.
"""
## @package view_polararray
# \ingroup draftviewproviders
# \brief Provides the viewprovider code for the polar Array object.

## \addtogroup draftviewproviders
# @{
import Draft_rc

from draftviewproviders.view_array import ViewProviderDraftArray

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class ViewProviderPolarArray(ViewProviderDraftArray):
    """View provider for the polar array object, currently unused."""

    def __init__(self, vobj):
        super().__init__(vobj)

    def getIcon(self):
        """Set the icon in the tree view."""
        return ":/icons/Draft_PolarArray"


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

from .asset import Asset
from .manager import AssetManager
from .uri import AssetUri
from .serializer import AssetSerializer, DummyAssetSerializer
from .store.base import AssetStore
from .store.memory import MemoryStore
from .store.filestore import FileStore

__all__ = [
    "Asset",
    "AssetUri",
    "AssetManager",
    "AssetSerializer",
    "DummyAssetSerializer",
    "AssetStore",
    "MemoryStore",
    "FileStore",
]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2018 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2020-2021 Schildkroet                                   *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Op.Base as PathOp
import Path.Op.EngraveBase as PathEngraveBase
import Path.Op.Util as PathOpUtil
import math

from PySide.QtCore import QT_TRANSLATE_NOOP

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Part = LazyLoader("Part", globals(), "Part")

__title__ = "CAM Deburr Operation"
__author__ = "sliptonic (Brad Collette), Schildkroet"
__url__ = "https://www.freecad.org"
__doc__ = "Deburr operation."


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


def toolDepthAndOffset(width, extraDepth, tool, printInfo):
    """toolDepthAndOffset(width, extraDepth, tool) ... return tuple for given\n
    parameters."""

    if not hasattr(tool, "Diameter"):
        raise ValueError("Deburr requires tool with diameter\n")

    suppressInfo = False
    if hasattr(tool, "CuttingEdgeAngle"):
        angle = float(tool.CuttingEdgeAngle)
        if Path.Geom.isRoughly(angle, 180) or Path.Geom.isRoughly(angle, 0):
            angle = 180
            toolOffset = float(tool.Diameter) / 2
        else:
            if hasattr(tool, "TipDiameter"):
                toolOffset = float(tool.TipDiameter) / 2
            elif hasattr(tool, "FlatRadius"):
                toolOffset = float(tool.FlatRadius)
            else:
                toolOffset = 0.0
                if printInfo and not suppressInfo:
                    FreeCAD.Console.PrintMessage(
                        translate(
                            "PathDeburr",
                            "The selected tool has no FlatRadius and no TipDiameter property. Assuming {}\n".format(
                                "Endmill" if angle == 180 else "V-Bit"
                            ),
                        )
                    )
                suppressInfo = True
    else:
        angle = 180
        toolOffset = float(tool.Diameter) / 2
        if printInfo:
            FreeCAD.Console.PrintMessage(
                translate(
                    "PathDeburr",
                    "The selected tool has no CuttingEdgeAngle property. Assuming Endmill\n",
                )
            )
        suppressInfo = True

    tan = math.tan(math.radians(angle / 2))

    toolDepth = 0 if Path.Geom.isRoughly(tan, 0) else width / tan
    depth = toolDepth + extraDepth
    extraOffset = -width if angle == 180 else (extraDepth * tan)
    offset = toolOffset + extraOffset

    return (depth, offset, extraOffset, suppressInfo)


class ObjectDeburr(PathEngraveBase.ObjectOp):
    """Proxy class for Deburr operation."""

    def opFeatures(self, obj):
        return (
            PathOp.FeatureTool
            | PathOp.FeatureHeights
            | PathOp.FeatureStepDown
            | PathOp.FeatureBaseEdges
            | PathOp.FeatureBaseFaces
            | PathOp.FeatureCoolant
            | PathOp.FeatureBaseGeometry
        )

    def initOperation(self, obj):
        Path.Log.track(obj.Label)
        obj.addProperty(
            "App::PropertyDistance",
            "Width",
            "Deburr",
            QT_TRANSLATE_NOOP("App::Property", "The desired width of the chamfer"),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "ExtraDepth",
            "Deburr",
            QT_TRANSLATE_NOOP("App::Property", "The additional depth of the toolpath"),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "Join",
            "Deburr",
            QT_TRANSLATE_NOOP("App::Property", "How to join chamfer segments"),
        )
        # obj.Join = ["Round", "Miter"]
        obj.setEditorMode("Join", 2)  # hide for now
        obj.addProperty(
            "App::PropertyEnumeration",
            "Direction",
            "Deburr",
            QT_TRANSLATE_NOOP("App::Property", "Direction of toolpath"),
        )
        # obj.Direction = ["CW", "CCW"]
        obj.addProperty(
            "App::PropertyEnumeration",
            "Side",
            "Deburr",
            QT_TRANSLATE_NOOP("App::Property", "Side of base object"),
        )
        obj.Side = ["Outside", "Inside"]
        obj.setEditorMode("Side", 2)  # Hide property, it's calculated by op
        obj.addProperty(
            "App::PropertyInteger",
            "EntryPoint",
            "Deburr",
            QT_TRANSLATE_NOOP("App::Property", "The segment where the toolpath starts"),
        )

        ENUMS = self.propertyEnumerations()
        for n in ENUMS:
            setattr(obj, n[0], n[1])

    @classmethod
    def propertyEnumerations(self, dataType="data"):
        """opPropertyEnumerations(dataType="data")... return property enumeration lists of specified dataType.
        Args:
            dataType = 'data', 'raw', 'translated'
        Notes:
        'data' is list of internal string literals used in code
        'raw' is list of (translated_text, data_string) tuples
        'translated' is list of translated string literals
        """

        # Enumeration lists for App::PropertyEnumeration properties
        enums = {
            "Direction": [
                (translate("Path", "CW"), "CW"),
                (translate("Path", "CCW"), "CCW"),
            ],  # this is the direction that the profile runs
            "Join": [
                (translate("PathDeburr", "Round"), "Round"),
                (translate("PathDeburr", "Miter"), "Miter"),
            ],  # this is the direction that the profile runs
        }

        if dataType == "raw":
            return enums

        data = list()
        idx = 0 if dataType == "translated" else 1

        Path.Log.debug(enums)

        for k, v in enumerate(enums):
            # data[k] = [tup[idx] for tup in v]
            data.append((v, [tup[idx] for tup in enums[v]]))
        Path.Log.debug(data)

        return data

    def opOnDocumentRestored(self, obj):
        obj.setEditorMode("Join", 2)  # hide for now

    def opExecute(self, obj):
        Path.Log.track(obj.Label)

        if not obj.Base:
            return

        if not hasattr(self, "printInfo"):
            self.printInfo = True
        try:
            (depth, offset, extraOffset, suppressInfo) = toolDepthAndOffset(
                obj.Width.Value, obj.ExtraDepth.Value, self.tool, self.printInfo
            )
            self.printInfo = not suppressInfo
        except ValueError as e:
            msg = "{} \n No path will be generated".format(e)
            raise ValueError(msg)
            # QtGui.QMessageBox.information(None, "Tool Error", msg)
            # return

        Path.Log.track(obj.Label, depth, offset)

        self.basewires = []
        self.adjusted_basewires = []
        wires = []

        for base, subs in obj.Base:
            edges = []
            basewires = []
            max_h = -99999
            radius_top = 0
            radius_bottom = 0

            for f in subs:
                sub = base.Shape.getElement(f)

                if type(sub) == Part.Edge:  # Edge
                    edges.append(sub)

                elif type(sub) == Part.Face and sub.normalAt(0, 0) != FreeCAD.Vector(
                    0, 0, 1
                ):  # Angled face
                    # If an angled face is selected, the lower edge is projected to the height of the upper edge,
                    # to simulate an edge

                    # Find z value of upper edge
                    for edge in sub.Edges:
                        for p0 in edge.Vertexes:
                            if p0.Point.z > max_h:
                                max_h = p0.Point.z

                    # Find biggest radius for top/bottom
                    for edge in sub.Edges:
                        if Part.Circle == type(edge.Curve):
                            if edge.Vertexes[0].Point.z == max_h:
                                if edge.Curve.Radius > radius_top:
                                    radius_top = edge.Curve.Radius
                            else:
                                if edge.Curve.Radius > radius_bottom:
                                    radius_bottom = edge.Curve.Radius

                    # Search for lower edge and raise it to height of upper edge
                    for edge in sub.Edges:
                        if Part.Circle == type(edge.Curve):  # Edge is a circle
                            if edge.Vertexes[0].Point.z < max_h:

                                if edge.Closed:  # Circle
                                    # New center
                                    center = FreeCAD.Vector(
                                        edge.Curve.Center.x, edge.Curve.Center.y, max_h
                                    )
                                    new_edge = Part.makeCircle(
                                        edge.Curve.Radius,
                                        center,
                                        FreeCAD.Vector(0, 0, 1),
                                    )
                                    edges.append(new_edge)

                                    # Modify offset for inner angled faces
                                    if radius_bottom < radius_top:
                                        offset -= 2 * extraOffset

                                    break

                                else:  # Arc
                                    if edge.Vertexes[0].Point.z == edge.Vertexes[1].Point.z:
                                        # Arc vertexes are on same layer
                                        l1 = math.sqrt(
                                            (edge.Vertexes[0].Point.x - edge.Curve.Center.x) ** 2
                                            + (edge.Vertexes[0].Point.y - edge.Curve.Center.y) ** 2
                                        )
                                        l2 = math.sqrt(
                                            (edge.Vertexes[1].Point.x - edge.Curve.Center.x) ** 2
                                            + (edge.Vertexes[1].Point.y - edge.Curve.Center.y) ** 2
                                        )

                                        # New center
                                        center = FreeCAD.Vector(
                                            edge.Curve.Center.x,
                                            edge.Curve.Center.y,
                                            max_h,
                                        )

                                        # Calculate angles based on x-axis (0 - PI/2)
                                        start_angle = math.acos(
                                            (edge.Vertexes[0].Point.x - edge.Curve.Center.x) / l1
                                        )
                                        end_angle = math.acos(
                                            (edge.Vertexes[1].Point.x - edge.Curve.Center.x) / l2
                                        )

                                        # Angles are based on x-axis (Mirrored on x-axis) -> negative y value means negative angle
                                        if edge.Vertexes[0].Point.y < edge.Curve.Center.y:
                                            start_angle *= -1
                                        if edge.Vertexes[1].Point.y < edge.Curve.Center.y:
                                            end_angle *= -1

                                        # Create new arc
                                        new_edge = Part.ArcOfCircle(
                                            Part.Circle(
                                                center,
                                                FreeCAD.Vector(0, 0, 1),
                                                edge.Curve.Radius,
                                            ),
                                            start_angle,
                                            end_angle,
                                        ).toShape()
                                        edges.append(new_edge)

                                        # Modify offset for inner angled faces
                                        if radius_bottom < radius_top:
                                            offset -= 2 * extraOffset

                                        break

                        else:  # Line
                            if (
                                edge.Vertexes[0].Point.z == edge.Vertexes[1].Point.z
                                and edge.Vertexes[0].Point.z < max_h
                            ):
                                new_edge = Part.Edge(
                                    Part.LineSegment(
                                        FreeCAD.Vector(
                                            edge.Vertexes[0].Point.x,
                                            edge.Vertexes[0].Point.y,
                                            max_h,
                                        ),
                                        FreeCAD.Vector(
                                            edge.Vertexes[1].Point.x,
                                            edge.Vertexes[1].Point.y,
                                            max_h,
                                        ),
                                    )
                                )
                                edges.append(new_edge)

                elif sub.Wires:
                    basewires.extend(sub.Wires)

                else:  # Flat face
                    basewires.append(Part.Wire(sub.Edges))

            self.edges = edges
            for edgelist in Part.sortEdges(edges):
                basewires.append(Part.Wire(edgelist))

            self.basewires.extend(basewires)

            # Set default side
            side = ["Outside"]

            for w in basewires:
                self.adjusted_basewires.append(w)
                wire = PathOpUtil.offsetWire(w, base.Shape, offset, True, side)
                if wire:
                    wires.append(wire)

        # Set direction of op
        forward = obj.Direction == "CW"

        # Set value of side
        obj.Side = side[0]
        # Check side extra for angled faces
        if radius_top > radius_bottom:
            obj.Side = "Inside"

        zValues = []
        z = 0
        if obj.StepDown.Value != 0:
            while z + obj.StepDown.Value < depth:
                z = z + obj.StepDown.Value
                zValues.append(z)

        zValues.append(depth)
        Path.Log.track(obj.Label, depth, zValues)

        if obj.EntryPoint < 0:
            obj.EntryPoint = 0

        self.wires = wires
        self.buildpathocc(obj, wires, zValues, True, forward, obj.EntryPoint)

    def opRejectAddBase(self, obj, base, sub):
        """The chamfer op can only deal with features of the base model, all others are rejected."""
        return base not in self.model

    def opSetDefaultValues(self, obj, job):
        Path.Log.track(obj.Label, job.Label)
        obj.Width = "1 mm"
        obj.ExtraDepth = "0.5 mm"
        obj.Join = "Round"
        obj.setExpression("StepDown", "0 mm")
        obj.StepDown = "0 mm"
        obj.Direction = "CW"
        obj.Side = "Outside"
        obj.EntryPoint = 0


def SetupProperties():
    setup = []
    setup.append("Width")
    setup.append("ExtraDepth")
    return setup


def Create(name, obj=None, parentJob=None):
    """Create(name) ... Creates and returns a Deburr operation."""
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    obj.Proxy = ObjectDeburr(obj, name, parentJob)
    return obj



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path
import Path.Base.Gui.Util as PathGuiUtil
import Path.Op.Gui.Base as PathOpGui
import Path.Op.Pocket as PathPocket
import PathGui

__title__ = "CAM Pocket Base Operation UI"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Base page controller and command implementation for pocket operations."

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate

FeaturePocket = 0x01
FeatureFacing = 0x02
FeatureOutline = 0x04
FeatureRestMachining = 0x08


class TaskPanelOpPage(PathOpGui.TaskPanelPage):
    """Page controller class for pocket operations, supports:
    FeaturePocket  ... used for pocketing operation
    FeatureFacing  ... used for face milling operation
    FeatureOutline ... used for pocket-shape operation
    """

    def pocketFeatures(self):
        """pocketFeatures() ... return which features of the UI are supported by the operation.
          FeaturePocket  ... used for pocketing operation
          FeatureFacing  ... used for face milling operation
          FeatureOutline ... used for pocket-shape operation
        Must be overwritten by subclasses"""
        pass

    def getForm(self):
        """getForm() ... returns UI, adapted to the results from pocketFeatures()"""
        form = FreeCADGui.PySideUic.loadUi(":/panels/PageOpPocketFullEdit.ui")

        comboToPropertyMap = [
            ("cutMode", "CutMode"),
            ("clearingPattern", "ClearingPattern"),
        ]
        enumTups = PathPocket.ObjectPocket.pocketPropertyEnumerations(dataType="raw")

        self.populateCombobox(form, enumTups, comboToPropertyMap)

        if not FeatureFacing & self.pocketFeatures():
            form.facingWidget.hide()
            form.clearEdges.hide()

        if FeaturePocket & self.pocketFeatures():
            form.extraOffset_label.setText(translate("PathPocket", "Pass Extension"))
            form.extraOffset.setToolTip(
                translate(
                    "PathPocket",
                    "The distance the facing operation will extend beyond the boundary shape.",
                )
            )

        if not (FeatureOutline & self.pocketFeatures()):
            form.useOutline.hide()

        if not (FeatureRestMachining & self.pocketFeatures()):
            form.useRestMachining.hide()

        return form

    def updateMinTravel(self, obj, setModel=True):
        if setModel and obj.MinTravel != self.form.minTravel.isChecked():
            obj.MinTravel = self.form.minTravel.isChecked()

    def updateZigZagAngle(self, obj, setModel=True):
        if obj.ClearingPattern in ["Offset"]:
            self.form.zigZagAngle.setEnabled(False)
        else:
            self.form.zigZagAngle.setEnabled(True)

        if setModel:
            PathGuiUtil.updateInputField(obj, "ZigZagAngle", self.form.zigZagAngle)

    def getFields(self, obj):
        """getFields(obj) ... transfers values from UI to obj's properties"""
        if obj.CutMode != str(self.form.cutMode.currentData()):
            obj.CutMode = str(self.form.cutMode.currentData())
        if obj.StepOver != self.form.stepOverPercent.value():
            obj.StepOver = self.form.stepOverPercent.value()
        if obj.ClearingPattern != str(self.form.clearingPattern.currentData()):
            obj.ClearingPattern = str(self.form.clearingPattern.currentData())

        PathGuiUtil.updateInputField(obj, "ExtraOffset", self.form.extraOffset)
        self.updateToolController(obj, self.form.toolController)
        self.updateCoolant(obj, self.form.coolantController)
        self.updateZigZagAngle(obj)

        if obj.UseStartPoint != self.form.useStartPoint.isChecked():
            obj.UseStartPoint = self.form.useStartPoint.isChecked()

        if obj.UseRestMachining != self.form.useRestMachining.isChecked():
            obj.UseRestMachining = self.form.useRestMachining.isChecked()

        if FeatureOutline & self.pocketFeatures():
            if obj.UseOutline != self.form.useOutline.isChecked():
                obj.UseOutline = self.form.useOutline.isChecked()

        self.updateMinTravel(obj)

        if FeatureFacing & self.pocketFeatures():
            print(obj.BoundaryShape)
            print(self.form.boundaryShape.currentText())
            print(self.form.boundaryShape.currentData())
            if obj.BoundaryShape != str(self.form.boundaryShape.currentData()):
                obj.BoundaryShape = str(self.form.boundaryShape.currentData())
            if obj.ClearEdges != self.form.clearEdges.isChecked():
                obj.ClearEdges = self.form.clearEdges.isChecked()

    def setFields(self, obj):
        """setFields(obj) ... transfers obj's property values to UI"""
        self.form.stepOverPercent.setValue(obj.StepOver)
        self.form.extraOffset.setText(
            FreeCAD.Units.Quantity(obj.ExtraOffset.Value, FreeCAD.Units.Length).UserString
        )
        self.form.useStartPoint.setChecked(obj.UseStartPoint)
        self.form.useRestMachining.setChecked(obj.UseRestMachining)
        if FeatureOutline & self.pocketFeatures():
            self.form.useOutline.setChecked(obj.UseOutline)

        self.form.zigZagAngle.setText(
            FreeCAD.Units.Quantity(obj.ZigZagAngle, FreeCAD.Units.Angle).UserString
        )
        self.updateZigZagAngle(obj, False)

        self.form.minTravel.setChecked(obj.MinTravel)
        self.updateMinTravel(obj, False)

        self.selectInComboBox(obj.ClearingPattern, self.form.clearingPattern)
        self.selectInComboBox(obj.CutMode, self.form.cutMode)
        self.setupToolController(obj, self.form.toolController)
        self.setupCoolant(obj, self.form.coolantController)

        if FeatureFacing & self.pocketFeatures():
            self.selectInComboBox(obj.BoundaryShape, self.form.boundaryShape)
            self.form.clearEdges.setChecked(obj.ClearEdges)

    def getSignalsForUpdate(self, obj):
        """getSignalsForUpdate(obj) ... return list of signals for updating obj"""
        signals = []

        signals.append(self.form.cutMode.currentIndexChanged)
        signals.append(self.form.clearingPattern.currentIndexChanged)
        signals.append(self.form.stepOverPercent.editingFinished)
        signals.append(self.form.zigZagAngle.editingFinished)
        signals.append(self.form.toolController.currentIndexChanged)
        signals.append(self.form.extraOffset.editingFinished)
        signals.append(self.form.useStartPoint.clicked)
        signals.append(self.form.useRestMachining.clicked)
        signals.append(self.form.useOutline.clicked)
        signals.append(self.form.minTravel.clicked)
        signals.append(self.form.coolantController.currentIndexChanged)

        if FeatureFacing & self.pocketFeatures():
            signals.append(self.form.boundaryShape.currentIndexChanged)
            signals.append(self.form.clearEdges.clicked)

        return signals
